import logging
from typing import Callable

from overhave.transport.redis.consumer import RedisConsumer
from overhave.transport.redis.objects import AnyRedisTask, RedisContainer, RedisUnreadData

logger = logging.getLogger(__name__)


class RedisConsumerRunnerException(Exception):
    """Exception for RedisConsumerRunner errors."""

    pass


class RedisConsumerRunner:
    """Class for running tasks specified by ```mapping```.

    Runner tasks launch with instance of :class:`RedisConsumer` ```consumer```.
    """

    def __init__(
        self, consumer: RedisConsumer, mapping: dict[type[AnyRedisTask], Callable[[AnyRedisTask], None]]
    ) -> None:
        self._consumer = consumer
        self._mapping = mapping

    def run(self) -> None:
        try:
            self._run()
        except Exception as e:
            raise RedisConsumerRunnerException from e

        logger.info("Shutdown event reached")

    def _run(self) -> None:
        with self._consumer:
            for message_sequence in self._consumer:
                for msg in message_sequence:
                    self._process(msg)

    def _process(self, data: RedisUnreadData) -> None:
        container = RedisContainer(task=data.decoded_message)
        logger.info("Gotten ready for test_execution BaseRedisTask: %s", container.task)
        self._mapping[type(container.task)](container.task)



import abc
from functools import cached_property

import sqlalchemy.orm as so

from overhave import db
from overhave.storage.converters import FeatureTypeModel, FeatureTypeName


class BaseFeatureTypeStorageException(Exception):
    """Base exception for :class:`FeatureTypeStorage`."""


class FeatureTypeNotExistsError(BaseFeatureTypeStorageException):
    """Exception for situation without specified feature type."""


class IFeatureTypeStorage(abc.ABC):
    """Abstract class for feature type storage."""

    @abc.abstractmethod
    @cached_property
    def default_feature_type_name(self) -> FeatureTypeName:
        pass

    @staticmethod
    @abc.abstractmethod
    def feature_type_by_name(session: so.Session, name: FeatureTypeName) -> db.FeatureType:
        pass

    @abc.abstractmethod
    @cached_property
    def all_feature_types(self) -> list[FeatureTypeModel]:
        pass


class FeatureTypeStorage(IFeatureTypeStorage):
    """Class for feature type storage."""

    @cached_property
    def default_feature_type_name(self) -> FeatureTypeName:
        with db.create_session() as session:
            row_entities: tuple[FeatureTypeName] | None = (
                session.query(db.FeatureType)
                .with_entities(db.FeatureType.name)
                .order_by(db.FeatureType.id.asc())
                .first()
            )
            if row_entities is None:
                raise FeatureTypeNotExistsError("No one feature type exists in db.FeatureType table!")
            return row_entities[0]

    @staticmethod
    def feature_type_by_name(session: so.Session, name: FeatureTypeName) -> db.FeatureType:
        feature_type = session.query(db.FeatureType).filter(db.FeatureType.name == name).one_or_none()
        if feature_type is None:
            raise FeatureTypeNotExistsError(f"Could not find feature type with name='{name}'!")
        return feature_type

    @cached_property
    def all_feature_types(self) -> list[FeatureTypeModel]:
        with db.create_session() as session:
            db_feature_types = session.query(db.FeatureType).all()
            return [FeatureTypeModel.model_validate(x) for x in db_feature_types]



import abc

import sqlalchemy.orm as so
from pydantic import SecretStr

from overhave import db
from overhave.storage.converters import SystemUserModel


class ISystemUserStorage(abc.ABC):
    """Abstract class for system user storage."""

    @staticmethod
    @abc.abstractmethod
    def get_user_model(user_id: int) -> SystemUserModel | None:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_user(
        session: so.Session, login: str, password: SecretStr | None = None, role: db.Role = db.Role.user
    ) -> db.UserRole:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_user_by_credits(session: so.Session, login: str, password: SecretStr | None = None) -> db.UserRole | None:
        pass


class SystemUserStorage(ISystemUserStorage):
    """Class for system user storage."""

    @staticmethod
    def get_user_model(user_id: int) -> SystemUserModel | None:
        with db.create_session() as session:
            db_user = session.get(db.UserRole, user_id)
            if db_user is not None:
                return SystemUserModel.model_validate(db_user)
            return None

    @staticmethod
    def create_user(
        session: so.Session, login: str, password: SecretStr | None = None, role: db.Role = db.Role.user
    ) -> db.UserRole:
        db_password = None
        if password is not None:
            db_password = password.get_secret_value()
        db_user = db.UserRole(login=login, password=db_password, role=role)
        session.add(db_user)
        session.flush()
        return db_user

    @staticmethod
    def get_user_by_credits(session: so.Session, login: str, password: SecretStr | None = None) -> db.UserRole | None:
        query = session.query(db.UserRole).filter(db.UserRole.login == login)
        if password is not None:
            query = query.filter(db.UserRole.password == password.get_secret_value())
        return query.one_or_none()



from flask import Flask


def get_flask_app(template_folder: str) -> Flask:
    app = Flask("admin", template_folder=template_folder)
    app.secret_key = "overhave cool secret key"
    app.config.update({"FLASK_ADMIN_SWATCH": "united", "FLASK_ADMIN_FLUID_LAYOUT": True})
    return app



# flake8: noqa
from .compiler import IncorrectScenarioTextError, OverhaveScenarioCompilerSettings, ScenarioCompiler, generate_task_info
from .file_manager import EmptyGitProjectURLError, EmptyTaskTrackerURLError, FileManager, OverhaveProjectSettings
from .parser import (
    FeatureInfo,
    FeatureNameParsingError,
    NullableFeatureIdError,
    OverhaveScenarioParserSettings,
    ScenarioParser,
    StrictFeatureInfo,
    StrictFeatureParsingError,
)
from .validator import FeatureValidator, IFeatureValidator



from overhave.transport.http import BaseHttpClientSettings
from overhave.transport.http.gitlab_client.objects import TokenType


class OverhaveGitlabClientSettings(BaseHttpClientSettings):
    """Settings for :class:`GitlabHttpClient`."""

    auth_token: str | None = None
    token_type: TokenType = TokenType.OAUTH

    class Config:
        env_prefix = "OVERHAVE_GITLAB_"



import abc
from typing import cast

import sqlalchemy as sa
import sqlalchemy.orm as so

from overhave import db
from overhave.storage.converters import ScenarioModel


class IScenarioStorage(abc.ABC):
    """Abstract class for feature type storage."""

    @staticmethod
    @abc.abstractmethod
    def scenario_model_by_id(session: so.Session, scenario_id: int) -> ScenarioModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_scenario_by_feature_id(feature_id: int) -> ScenarioModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_scenario(model: ScenarioModel) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_scenario(model: ScenarioModel) -> int:
        pass


class ScenarioStorage(IScenarioStorage):
    """Class for feature type storage."""

    @staticmethod
    def scenario_model_by_id(session: so.Session, scenario_id: int) -> ScenarioModel:
        scenario = session.query(db.Scenario).filter(db.Scenario.id == scenario_id).one()
        return ScenarioModel.model_validate(scenario)

    @staticmethod
    def get_scenario_by_feature_id(feature_id: int) -> ScenarioModel:
        with db.create_session() as session:
            scenario: db.Scenario = session.query(db.Scenario).filter(db.Scenario.feature_id == feature_id).one()
            return ScenarioModel.model_validate(scenario)

    @staticmethod
    def update_scenario(model: ScenarioModel) -> None:
        with db.create_session() as session:
            session.execute(sa.update(db.Scenario).where(db.Scenario.id == model.id).values(text=model.text))

    @staticmethod
    def create_scenario(model: ScenarioModel) -> int:
        with db.create_session() as session:
            scenario = db.Scenario(feature_id=model.feature_id, text=model.text)
            session.add(scenario)
            session.flush()
            return cast(int, scenario.id)



import typer

from overhave.base_settings import DataBaseSettings, LoggingSettings
from overhave.factory import get_synchronizer_factory
from overhave.scenario import IFeatureValidator
from overhave.synchronization import IOverhaveSynchronizer

sync_app = typer.Typer(short_help="Run Overhave features synchronization commands")


def _create_synchronizer() -> IOverhaveSynchronizer:
    DataBaseSettings().setup_engine()
    LoggingSettings().setup_logging()
    return get_synchronizer_factory().synchronizer


@sync_app.command(short_help="Run Overhave features synchronization")
def run(
    create_db_features: bool = typer.Option(
        False, "-c", "--create-db-features", is_flag=True, help="Create features in database if necessary"
    ),
    pull_repository: bool = typer.Option(False, "-p", "--pull-repository", is_flag=True, help="Pull remote repository"),
) -> None:
    _create_synchronizer().synchronize(create_db_features, pull_repository)


def _create_validator() -> IFeatureValidator:
    LoggingSettings().setup_logging()
    return get_synchronizer_factory().feature_validator


@sync_app.command(short_help="Run Overhave feature files validation")
def validate_features(
    raise_if_nullable_id: bool = typer.Option(
        False, "-r", "--raise-if-nullable-id", is_flag=True, help="Raise if validator find features with nullable IDs"
    ),
    pull_repository: bool = typer.Option(False, "-p", "--pull-repository", is_flag=True, help="Pull remote repository"),
) -> None:
    _create_validator().validate(raise_if_nullable_id, pull_repository)



import logging
from typing import Any, Callable

import _pytest
import allure
import httpx
from _pytest.config import Config
from _pytest.fixtures import FixtureRequest
from _pytest.main import Session
from _pytest.nodes import Item
from _pytest.python import Function
from pydantic import ValidationError
from pytest_bdd.parser import Feature, Scenario, Step

from overhave.factory import IAdminFactory
from overhave.pytest_plugin.deps import get_description_manager, get_step_context_runner, get_tag_controller
from overhave.pytest_plugin.helpers import (
    add_admin_feature_link_to_report,
    add_scenario_title_to_report,
    add_task_links_to_report,
    get_feature_info_from_item,
    get_full_step_name,
    is_pytest_bdd_item,
    set_feature_info_for_item,
    set_git_project_url_if_necessary,
    set_severity_level,
)
from overhave.pytest_plugin.proxy_manager import get_proxy_manager

logger = logging.getLogger(__name__)


class StepNotFoundError(RuntimeError):
    """Exception for situation with missing or incorrect step definition."""


def pytest_configure(config: Config) -> None:
    """Patch pytest_bdd objects in current hook."""
    proxy_manager = get_proxy_manager()
    if not proxy_manager.has_factory:
        logger.debug("Overhave ProxyManager has not got prepared factory, so skip injection.")
        return
    tw = _pytest.config.create_terminal_writer(config)
    try:
        logger.debug("Try to patch pytest objects...")
        proxy_manager.patch_pytest()
        logger.debug("Successfully patched pytest objects.")
        tw.line("Overhave injector successfully initialized.", green=True)
    except ValidationError as e:
        tw.line(f"Could not initialize Overhave injector!\n{str(e)}", red=True)


def pytest_collection_modifyitems(session: Session) -> None:
    pytest_bdd_scenario_items = (item for item in session.items if is_pytest_bdd_item(item))
    for item in pytest_bdd_scenario_items:
        add_scenario_title_to_report(item)


def pytest_bdd_before_step(
    request: FixtureRequest, feature: Feature, scenario: Scenario, step: Step, step_func: Callable[[Any], None]
) -> None:
    get_step_context_runner.cache_clear()
    runner = get_step_context_runner()
    runner.set_title(get_full_step_name(step))
    runner.start()


def pytest_bdd_after_step(
    request: FixtureRequest,
    feature: Feature,
    scenario: Scenario,
    step: Step,
    step_func: Callable[[Any], None],
    step_func_args: dict[str, Any],
) -> None:
    runner = get_step_context_runner()
    runner.stop(None)


def pytest_bdd_step_error(
    request: FixtureRequest,
    feature: Feature,
    scenario: Scenario,
    step: Step,
    step_func: Callable[[Any], None],
    step_func_args: dict[str, Any],
    exception: BaseException,
) -> None:
    runner = get_step_context_runner()
    runner.stop(exception)


def pytest_bdd_apply_tag(tag: str, function: Function) -> bool | None:
    controller = get_tag_controller()
    tag_pattern = controller.get_suitable_pattern(tag)
    if not tag_pattern:
        logger.debug("Tag '%s' is not suitable for evaluation", tag)
        return None
    result = controller.evaluate_tag(name=tag, pattern=tag_pattern)
    result.marker(function)
    if result.url is not None:
        allure.dynamic.link(url=result.url, link_type=result.link_type, name=result.url)
    logger.debug("Tag '%s' successfully evaluated")
    return True


def pytest_bdd_step_func_lookup_error(
    request: FixtureRequest, feature: Feature, scenario: Scenario, step: Step, exception: BaseException
) -> None:
    raise StepNotFoundError(f"Could not found specified step '{get_full_step_name(step)}'") from exception


def pytest_collection_finish(session: Session) -> None:
    """Supplying of injector configs for steps collection (only :class:`IAdminFactory`)."""
    proxy_manager = get_proxy_manager()
    if not proxy_manager.has_factory or not isinstance(proxy_manager.factory, IAdminFactory):
        return
    tw = _pytest.config.create_terminal_writer(session.config)
    if not proxy_manager.pytest_patched:
        tw.line("Could not supplement Overhave injector - pytest session has not been patched!", yellow=True)
        return
    try:
        proxy_manager.supply_injector_for_collection()
        tw.line("Overhave injector successfully supplemented.", green=True)
    except ValidationError as e:
        tw.line(f"Could not supplement Overhave injector!\n{str(e)}", red=True)
    proxy_manager.injector.adapt(session)


def pytest_runtest_setup(item: Item) -> None:
    """Hook for purgation of `get_description_manager` func and upgrading item for reports."""
    get_description_manager.cache_clear()
    proxy_manager = get_proxy_manager()
    if not proxy_manager.has_factory or not is_pytest_bdd_item(item):
        return

    set_feature_info_for_item(item=item, scenario_parser=proxy_manager.factory.scenario_parser)
    feature_info = get_feature_info_from_item(item)

    admin_link_settings = proxy_manager.factory.context.admin_link_settings  # type: ignore
    if admin_link_settings.enabled and feature_info.id is not None:
        add_admin_feature_link_to_report(admin_link_settings=admin_link_settings, feature_id=feature_info.id)

    project_settings = proxy_manager.factory.context.project_settings
    if isinstance(project_settings.git_project_url, httpx.URL) and feature_info.type is not None:
        set_git_project_url_if_necessary(
            project_settings=project_settings,
            feature_extractor=proxy_manager.factory.feature_extractor,
            item=item,
            feature_type=feature_info.type,
        )
    if isinstance(project_settings.tasks_keyword, str) and feature_info.tasks:
        add_task_links_to_report(project_settings=project_settings, tasks=feature_info.tasks)

    set_severity_level(
        compilation_settings=proxy_manager.factory.context.compilation_settings,
        item=item,
    )


def pytest_runtest_teardown(item: Item, nextitem: Item | None) -> None:
    """Hook for description attachment to Allure report."""
    if not get_proxy_manager().has_factory:
        return
    get_description_manager().apply_description()



from dataclasses import dataclass, field
from datetime import datetime

import allure

from overhave.storage import FeatureTypeName


@dataclass
class FeatureInfo:
    """Model for feature info keeping."""

    id: int | None = None
    name: str | None = None
    type: FeatureTypeName | None = None
    tags: list[str] | None = None
    severity: allure.severity_level | None = None
    author: str | None = None
    last_edited_by: str | None = None
    last_edited_at: datetime | None = None
    tasks: list[str] | None = None
    scenarios: str | None = None


@dataclass
class StrictFeatureInfo:
    """Model for feature info keeping with strict presence of fields."""

    id: int
    name: str
    type: FeatureTypeName
    author: str
    last_edited_by: str
    last_edited_at: datetime
    scenarios: str

    tags: list[str] = field(default_factory=list)
    severity: allure.severity_level = allure.severity_level.NORMAL
    tasks: list[str] = field(default_factory=list)



import logging
from datetime import datetime

import sqlalchemy.orm as so

from overhave.scenario import StrictFeatureInfo
from overhave.storage import (
    FeatureModel,
    FeatureTypeModel,
    FeatureTypeName,
    IDraftStorage,
    IFeatureStorage,
    IFeatureTagStorage,
    IFeatureTypeStorage,
    IScenarioStorage,
    ISystemUserStorage,
    ScenarioModel,
    TagModel,
)
from overhave.utils import ANY_INT

logger = logging.getLogger(__name__)


class BaseSynchronizerStorageMixinException(Exception):
    """Base exception for :class:`SynchronizerStorageManager`."""


class FeatureInfoUserNotFoundError(BaseSynchronizerStorageMixinException):
    """Exception for situation without specified user in database."""


class SynchronizerStorageManager:
    """Class for storages management while synchronizing features."""

    def __init__(
        self,
        feature_storage: IFeatureStorage,
        feature_type_storage: IFeatureTypeStorage,
        scenario_storage: IScenarioStorage,
        tag_storage: IFeatureTagStorage,
        draft_storage: IDraftStorage,
        system_user_storage: ISystemUserStorage,
    ) -> None:
        self._feature_storage = feature_storage
        self._feature_type_storage = feature_type_storage
        self._scenario_storage = scenario_storage
        self._tag_storage = tag_storage
        self._draft_storage = draft_storage
        self._system_user_storage = system_user_storage

    def get_feature(self, feature_id: int) -> FeatureModel | None:
        return self._feature_storage.get_feature_model(feature_id)

    def get_last_change_time(self, model: FeatureModel) -> datetime:
        last_draft_published_at = self._draft_storage.get_last_published_at_for_feature(model.id)
        if last_draft_published_at is not None:
            logger.info(
                "Last version has been published at %s.",
                last_draft_published_at.strftime("%d-%m-%Y %H:%M:%S"),
            )
            return last_draft_published_at
        logger.info("Feature hasn't got any published version.")
        return model.last_edited_at

    def get_feature_tags(self, session: so.Session, info: StrictFeatureInfo) -> list[TagModel]:
        db_tags = [
            self._tag_storage.get_or_create_tag(session=session, value=tag, created_by=info.last_edited_by)
            for tag in info.tags
        ]
        session.flush()
        return [TagModel.model_validate(x) for x in db_tags]

    def feature_type_by_name(self, session: so.Session, feature_type: FeatureTypeName) -> FeatureTypeModel:
        db_feature_type = self._feature_type_storage.feature_type_by_name(session=session, name=feature_type)
        return FeatureTypeModel.model_validate(db_feature_type)

    def ensure_users_exist(self, session: so.Session, info: StrictFeatureInfo) -> None:
        for user in {info.author, info.last_edited_by}:
            if self._system_user_storage.get_user_by_credits(session=session, login=user):
                continue
            raise FeatureInfoUserNotFoundError(f"Could not find user with login={user}!")

    def update_db_feature(self, model: FeatureModel, scenario: str) -> None:
        scenario_model = self._scenario_storage.get_scenario_by_feature_id(model.id)
        self._feature_storage.update_feature(model)
        scenario_model.text = scenario
        self._scenario_storage.update_scenario(model=scenario_model)

    def create_db_feature(self, model: FeatureModel, scenario: str) -> None:
        model.id = self._feature_storage.create_feature(model)
        scenario_model = ScenarioModel(id=ANY_INT, feature_id=model.id, text=scenario)
        self._scenario_storage.create_scenario(scenario_model)



import abc
from typing import Any, cast

import sqlalchemy as sa
import sqlalchemy.orm as so

from overhave import db
from overhave.storage.converters import TestRunModel
from overhave.utils import get_current_time


class ITestRunStorage(abc.ABC):
    """Abstract class for test runs storage."""

    @abc.abstractmethod
    def create_testrun(self, scenario_id: int, executed_by: str) -> int:
        pass

    @abc.abstractmethod
    def set_run_status(self, run_id: int, status: db.TestRunStatus, traceback: str | None = None) -> None:
        pass

    @abc.abstractmethod
    def set_report(self, run_id: int, status: db.TestReportStatus, report: str | None = None) -> None:
        pass

    @abc.abstractmethod
    def testrun_model_by_id(self, session: so.Session, run_id: int) -> TestRunModel:
        pass

    @abc.abstractmethod
    def get_testrun_model(self, run_id: int) -> TestRunModel | None:
        pass


class TestRunStorage(ITestRunStorage):
    """Class for test runs storage."""

    def create_testrun(self, scenario_id: int, executed_by: str) -> int:
        with db.create_session() as session:
            scenario: db.Scenario = session.query(db.Scenario).filter(db.Scenario.id == scenario_id).one()
            run = db.TestRun(
                scenario_id=scenario_id,
                name=scenario.feature.name,
                status=db.TestRunStatus.STARTED,
                report_status=db.TestReportStatus.EMPTY,
                executed_by=executed_by,
            )
            session.add(run)
            session.flush()
            return cast(int, run.id)

    def set_run_status(self, run_id: int, status: db.TestRunStatus, traceback: str | None = None) -> None:
        with db.create_session() as session:
            values: dict[str, Any] = {"status": status}
            if status == db.TestRunStatus.RUNNING:
                values["start"] = get_current_time()
            if status.finished:
                values["end"] = get_current_time()
            if isinstance(traceback, str):
                values["traceback"] = traceback

            session.execute(sa.update(db.TestRun).where(db.TestRun.id == run_id).values(**values))

    def set_report(self, run_id: int, status: db.TestReportStatus, report: str | None = None) -> None:
        with db.create_session() as session:
            values: dict[str, Any] = {"report_status": status}
            if isinstance(report, str):
                values["report"] = report

            session.execute(sa.update(db.TestRun).where(db.TestRun.id == run_id).values(**values))

    def testrun_model_by_id(self, session: so.Session, run_id: int) -> TestRunModel:
        run = session.query(db.TestRun).filter(db.TestRun.id == run_id).one()
        return TestRunModel.model_validate(run)

    def get_testrun_model(self, run_id: int) -> TestRunModel | None:
        with db.create_session() as session:
            run = session.get(db.TestRun, run_id)
            if run is None:
                return None
            return TestRunModel.model_validate(run)



import logging
from types import FunctionType
from typing import Any, Callable, Sequence, Type

from flask_admin.contrib.sqla import ModelView
from markupsafe import Markup

from overhave import db

logger = logging.getLogger(__name__)


def _default_formatter_result(
    model: db.BaseTable, name: str, supported_models: Sequence[type[db.BaseTable]]
) -> Markup | Any:
    value = getattr(model, name, None)
    if value is None:
        return Markup("")
    if not isinstance(model, tuple(supported_models)):
        return Markup(value)
    return value


def safe_formatter(type: Type[object], supported_models: Sequence[Type[db.BaseTable]]):  # type: ignore  # noqa: A002
    def decorator(func: FunctionType) -> Callable[[ModelView, Any, db.BaseTable, str], Markup | Any]:
        def wrapper(view: ModelView, context: Any, model: db.BaseTable, name: str) -> Markup | Any:
            logger.debug("Wrapping function '%s'...", func.__name__)
            result = _default_formatter_result(model=model, name=name, supported_models=supported_models)
            logger.debug("Default formatter result: '%s'", result)
            if not isinstance(result, Markup) and isinstance(result, type):
                result = func(view, context, model, result)
                logger.debug("'%s' result: '%s'", func.__name__, result)
            return result

        return wrapper

    return decorator



import abc
import logging
from typing import Iterable, cast

import sqlalchemy as sa
import sqlalchemy.orm as so

from overhave import db
from overhave.storage.converters import FeatureModel
from overhave.utils import get_current_time

logger = logging.getLogger(__name__)


class BaseFeatureStorageException(Exception):
    """Base exception for :class:`FeatureStorage`."""


class FeatureTagNotExistsError(BaseFeatureStorageException):
    """Error for situation when tag not found."""


class IFeatureStorage(abc.ABC):
    """Abstract class for feature storage."""

    @staticmethod
    @abc.abstractmethod
    def feature_model_by_id(session: so.Session, feature_id: int) -> FeatureModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_feature_model(feature_id: int) -> FeatureModel | None:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_feature(model: FeatureModel) -> int:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_feature(model: FeatureModel) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_features_by_tag(tag_id: int) -> list[FeatureModel]:
        pass


def _append_tags_to_feature(session: so.Session, feature: db.Feature, tag_ids: Iterable[int]) -> None:
    db_tags: list[db.Tags] = []
    for tag_id in tag_ids:
        db_tag: db.Tags | None = session.get(db.Tags, tag_id)
        if db_tag is None:
            raise FeatureTagNotExistsError(f"Feature tag with id={tag_id} does not exist!")
        logger.info("Append tag with id=%s and value=%s", tag_id, db_tag.value)
        db_tags.append(db_tag)
    feature.feature_tags.extend(db_tags)


class FeatureStorage(IFeatureStorage):
    """Class for feature storage."""

    @staticmethod
    def feature_model_by_id(session: so.Session, feature_id: int) -> FeatureModel:
        feature = session.query(db.Feature).filter(db.Feature.id == feature_id).one()
        return FeatureModel.model_validate(feature)

    @staticmethod
    def create_feature(model: FeatureModel) -> int:
        with db.create_session() as session:
            feature = db.Feature(
                name=model.name,
                author=model.author,
                type_id=model.feature_type.id,
                file_path=model.file_path,
                task=model.task,
                severity=model.severity,
                last_edited_by=model.last_edited_by,
                last_edited_at=model.last_edited_at,
                released=model.released,
            )
            _append_tags_to_feature(session=session, feature=feature, tag_ids=(tag.id for tag in model.feature_tags))
            session.add(feature)
            session.flush()
            return cast(int, feature.id)

    @staticmethod
    def update_feature(model: FeatureModel) -> None:
        with db.create_session() as session:
            session.execute(
                sa.update(db.Feature)
                .where(db.Feature.id == model.id)
                .values(
                    name=model.name,
                    file_path=model.file_path,
                    task=model.task,
                    severity=model.severity,
                    last_edited_by=model.last_edited_by,
                    last_edited_at=get_current_time(),
                    released=True,
                )
            )
            feature = session.query(db.Feature).filter(db.Feature.id == model.id).one()
            existing_tags = {tag.id for tag in feature.feature_tags}
            model_tags = {tag.id for tag in model.feature_tags}
            tags_to_delete = existing_tags.difference(model_tags)
            for tag_id in tags_to_delete:
                db_tag = next(tag for tag in feature.feature_tags if tag.id == tag_id)
                logger.info("Remove tag with id=%s and value=%s", tag_id, db_tag.value)
                feature.feature_tags.remove(db_tag)
            tags_to_append = model_tags.difference(existing_tags)
            _append_tags_to_feature(session=session, feature=feature, tag_ids=tags_to_append)

    @staticmethod
    def get_features_by_tag(tag_id: int) -> list[FeatureModel]:
        with db.create_session() as session:
            feature_ids_query = (
                session.query(db.FeatureTagsAssociationTable)
                .with_entities(db.FeatureTagsAssociationTable.feature_id)
                .filter(db.FeatureTagsAssociationTable.tags_id == tag_id)
                .scalar_subquery()
            )
            features = session.query(db.Feature).filter(db.Feature.id.in_(feature_ids_query)).all()
            return [FeatureModel.model_validate(x) for x in features]

    @staticmethod
    def get_feature_model(feature_id: int) -> FeatureModel | None:
        with db.create_session() as session:
            feature = session.get(db.Feature, feature_id)
            if feature is not None:
                return FeatureModel.model_validate(feature)
            return None



import abc
from pathlib import Path

from overhave.storage import FeatureTypeName


class IFeatureExtractor(abc.ABC):
    """Abstract class for specified project's feature types resolution."""

    @property
    @abc.abstractmethod
    def feature_types(self) -> list[FeatureTypeName]:
        pass

    @property
    @abc.abstractmethod
    def feature_type_to_dir_mapping(self) -> dict[FeatureTypeName, Path]:
        pass



import logging
from pathlib import Path

import pytest

from overhave.test_execution.settings import OverhaveTestSettings

logger = logging.getLogger(__name__)


def _extend_cmd_args(cmd: list[str], addoptions: str | None) -> None:
    if not isinstance(addoptions, str):
        return
    cmd.extend(addoptions.split(" "))


class PytestRunner:
    """Class for running `PyTest` in test and collect-only modes."""

    def __init__(self, settings: OverhaveTestSettings) -> None:
        self._settings = settings

    def run(self, fixture_file: str, alluredir: str) -> int:
        pytest_cmd = [fixture_file, f"--alluredir={alluredir}"]
        for addoptions in (self._settings.default_pytest_addoptions, self._settings.extra_pytest_addoptions):
            _extend_cmd_args(cmd=pytest_cmd, addoptions=addoptions)
        if self._settings.workers is not None:
            pytest_cmd.extend(["-n", f"{self._settings.workers}"])

        logger.debug("Prepared pytest args: %s", pytest_cmd)
        return pytest.main(pytest_cmd)

    def collect_only(self, fixture_file: Path) -> None:
        logger.info("Started tests collection process with '%s'...", fixture_file.name)
        pytest_cmd = [  # pylint: disable=W8301
            fixture_file.as_posix(),
            "--collect-only",
            "--disable-pytest-warnings",
        ]
        _extend_cmd_args(cmd=pytest_cmd, addoptions=self._settings.extra_pytest_addoptions)
        logger.debug("Prepared pytest args: %s", pytest_cmd)
        pytest.main(pytest_cmd)
        logger.info("Finished tests collection process with '%s'.", fixture_file.name)



class BaseGitVersionPublisherError(Exception):
    """Base exception for git version publisher."""


class CurrentBranchEqualToDefaultError(BaseGitVersionPublisherError):
    """Exception for situation when current branch is equal to default branch."""


class CurrentBranchNotEqualToTargetError(BaseGitVersionPublisherError):
    """Exception for situation when current branch is not equal to target branch."""


class CommitNotCreatedError(BaseGitVersionPublisherError):
    """Exception for situation with not created commit."""



import abc
import logging

import git

from overhave import db
from overhave.entities import GitPullError
from overhave.metrics import PublicationOverhaveMetricContainer
from overhave.publication.abstract_publisher import IVersionPublisher
from overhave.publication.errors import BaseGitVersionPublisherError
from overhave.scenario import FileManager, OverhaveProjectSettings, generate_task_info
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage, PublisherContext
from overhave.transport import PublicationTask

logger = logging.getLogger(__name__)


class BaseVersionPublisher(IVersionPublisher, abc.ABC):
    """Class for feature version's pull requests management relative to Atlassian Stash API."""

    def __init__(
        self,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
        metric_container: PublicationOverhaveMetricContainer,
    ) -> None:
        self._project_settings = project_settings
        self._feature_storage = feature_storage
        self._scenario_storage = scenario_storage
        self._test_run_storage = test_run_storage
        self._draft_storage = draft_storage
        self._file_manager = file_manager
        self._metric_container = metric_container

    def process_publish_task(self, task: PublicationTask) -> None:
        status = self.publish_version(task.data.draft_id)
        self._metric_container.add_publication_task_status(status=status.value)

    def _compile_context(self, draft_id: int) -> PublisherContext:
        with db.create_session() as session:
            draft_model = self._draft_storage.draft_model_by_id(session=session, draft_id=draft_id)
            test_run_model = self._test_run_storage.testrun_model_by_id(session=session, run_id=draft_model.test_run_id)
            feature_model = self._feature_storage.feature_model_by_id(
                session=session, feature_id=draft_model.feature_id
            )
            scenario_model = self._scenario_storage.scenario_model_by_id(
                session=session, scenario_id=test_run_model.scenario_id
            )
        return PublisherContext(
            feature=feature_model,
            scenario=scenario_model,
            test_run=test_run_model,
            draft=draft_model,
            target_branch=f"bdd-feature-{feature_model.id}",
        )

    def _compile_publication_description(self, context: PublisherContext) -> str:
        return "\n".join(
            (
                f"Feature ID: {context.feature.id}. Type: '{context.feature.feature_type.name}'.",
                f"Created by: @{context.feature.author} at {context.feature.created_at.strftime('%d-%m-%Y %H:%M:%S')}.",
                f"Last edited by: @{context.feature.last_edited_by}.",
                f"PR from Test Run ID: {context.test_run.id}. Executed by: @{context.test_run.executed_by}",
                f"Published by: @{context.draft.published_by}.",
                generate_task_info(tasks=context.feature.task, header=self._project_settings.tasks_keyword),
            )
        )

    @abc.abstractmethod
    def _prepare_repo(self, context: PublisherContext) -> None:
        pass

    def _prepare_publisher_context(self, draft_id: int) -> PublisherContext | None:
        logger.info("Start processing draft_id=%s...", draft_id)
        self._draft_storage.set_draft_status(draft_id, db.DraftStatus.CREATING)
        context = self._compile_context(draft_id)
        try:
            self._prepare_repo(context)
            return context
        except (git.GitCommandError, GitPullError, BaseGitVersionPublisherError) as err:
            logger.exception("Error while trying to pull or push!")
            self._draft_storage.set_draft_status(draft_id, db.DraftStatus.INTERNAL_ERROR, traceback=str(err))
            return None



from http import HTTPStatus

import fastapi
from fastapi import security as fastapi_security
from pydantic.types import SecretStr

from overhave import db
from overhave.api.auth import AUTH_HEADERS, create_access_token
from overhave.api.deps import get_api_auth_settings, get_system_user_storage
from overhave.api.settings import OverhaveApiAuthSettings
from overhave.storage import AuthToken, ISystemUserStorage
from overhave.utils import get_current_time


def login_for_access_token(
    form_data: fastapi_security.OAuth2PasswordRequestForm = fastapi.Depends(),
    auth_settings: OverhaveApiAuthSettings = fastapi.Depends(get_api_auth_settings),
    storage: ISystemUserStorage = fastapi.Depends(get_system_user_storage),
) -> AuthToken:
    with db.create_session() as session:
        if not storage.get_user_by_credits(
            session=session, login=form_data.username, password=SecretStr(form_data.password)
        ):
            raise fastapi.HTTPException(
                status_code=HTTPStatus.UNAUTHORIZED,
                detail="Incorrect username or password",
                headers=AUTH_HEADERS.model_dump(),
            )
    expires_at = get_current_time() + auth_settings.access_token_expire_timedelta
    access_token = create_access_token(auth_settings=auth_settings, username=form_data.username, expires_at=expires_at)
    return AuthToken(access_token=access_token, expires_at=expires_at)



import abc
from functools import cached_property

from overhave.emulation import Emulator
from overhave.factory.base_factory import BaseOverhaveFactory, IOverhaveFactory
from overhave.factory.components.abstract_consumer import ITaskConsumerFactory
from overhave.factory.context import OverhaveEmulationContext
from overhave.metrics import EmulationRunOverhaveMetricContainer, get_emulation_metric_container
from overhave.transport import EmulationTask


class IEmulationFactory(IOverhaveFactory[OverhaveEmulationContext], ITaskConsumerFactory[EmulationTask], abc.ABC):
    """Abstract factory for Overhave emulation application."""


class EmulationFactory(BaseOverhaveFactory[OverhaveEmulationContext], IEmulationFactory):
    """Factory for Overhave emulation application."""

    context_cls = OverhaveEmulationContext

    @cached_property
    def _emulator(self) -> Emulator:
        return Emulator(
            settings=self.context.emulation_settings,
            storage=self._emulation_storage,
            metric_container=self.metric_container,
        )

    def process_task(self, task: EmulationTask) -> None:
        return self._emulator.start_emulation(task)

    @property
    def metric_container(self) -> EmulationRunOverhaveMetricContainer:
        return get_emulation_metric_container()



import abc
import logging
import tempfile
from pathlib import Path

from overhave import db
from overhave.db import TestRunStatus
from overhave.entities import OverhaveFileSettings, ReportManager
from overhave.metrics import TestRunOverhaveMetricContainer
from overhave.scenario import FileManager
from overhave.storage import IFeatureStorage, IScenarioStorage, ITestRunStorage, TestExecutorContext
from overhave.test_execution.test_runner import PytestRunner
from overhave.transport import TestRunTask

logger = logging.getLogger(__name__)


class ITestExecutor(abc.ABC):
    """Abstract class for test execution."""

    @abc.abstractmethod
    def execute_test(self, test_run_id: int) -> None:
        pass

    @abc.abstractmethod
    def process_test_task(self, task: TestRunTask) -> None:
        pass


class TestExecutor(ITestExecutor):
    """Class for test execution."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        file_manager: FileManager,
        test_runner: PytestRunner,
        report_manager: ReportManager,
        metric_container: TestRunOverhaveMetricContainer,
    ):
        self._file_settings = file_settings
        self._feature_storage = feature_storage
        self._scenario_storage = scenario_storage
        self._test_run_storage = test_run_storage
        self._file_manager = file_manager
        self._test_runner = test_runner
        self._report_manager = report_manager
        self._metric_container = metric_container

    def _run_test(self, context: TestExecutorContext, alluredir: Path) -> int:
        with self._file_manager.tmp_feature_file(context=context) as feature_file:
            with self._file_manager.tmp_fixture_file(context=context, feature_file=feature_file) as fixture_file:
                return self._test_runner.run(fixture_file=fixture_file.name, alluredir=alluredir.as_posix())

    def _compile_context(self, test_run_id: int) -> TestExecutorContext:
        with db.create_session() as session:
            test_run_model = self._test_run_storage.testrun_model_by_id(session=session, run_id=test_run_id)
            scenario_model = self._scenario_storage.scenario_model_by_id(
                session=session, scenario_id=test_run_model.scenario_id
            )
            feature_model = self._feature_storage.feature_model_by_id(
                session=session, feature_id=scenario_model.feature_id
            )
        return TestExecutorContext(
            feature=feature_model,
            scenario=scenario_model,
            test_run=test_run_model,
        )

    def execute_test(self, test_run_id: int) -> None:
        self._test_run_storage.set_run_status(run_id=test_run_id, status=TestRunStatus.RUNNING)
        ctx = self._compile_context(test_run_id)

        results_dir = Path(tempfile.mkdtemp())
        logger.debug("Allure results directory path: %s", results_dir.as_posix())
        try:
            test_return_code = self._run_test(context=ctx, alluredir=results_dir)
        except Exception as e:
            logger.exception("Error!")
            self._test_run_storage.set_run_status(
                run_id=test_run_id, status=TestRunStatus.INTERNAL_ERROR, traceback=str(e)
            )
            self._metric_container.add_test_run_status(status=TestRunStatus.INTERNAL_ERROR.value)
            return

        logger.debug("Test returncode: %s", test_return_code)
        if test_return_code == 0:
            self._test_run_storage.set_run_status(run_id=test_run_id, status=TestRunStatus.SUCCESS)
            self._metric_container.add_test_run_status(status=TestRunStatus.SUCCESS.value)
        else:
            self._test_run_storage.set_run_status(
                run_id=test_run_id, status=TestRunStatus.FAILED, traceback="Test run failed!"
            )
            self._metric_container.add_test_run_status(status=TestRunStatus.FAILED.value)
        self._report_manager.create_allure_report(test_run_id=test_run_id, results_dir=results_dir)

    def process_test_task(self, task: TestRunTask) -> None:
        self.execute_test(test_run_id=task.data.test_run_id)



import re
from types import FunctionType
from typing import NewType

from pydantic import BaseModel, field_validator
from pytest_bdd.types import STEP_TYPES

StepTypeName = NewType("StepTypeName", str)
PUBLIC_STEP_ATTR_NAME = "_is_public_step"


class BddStepModel(BaseModel):
    """Model for pytest_bdd step in Overhave."""

    type: StepTypeName
    name: str
    doc: str

    @field_validator("type")
    def validate_type(cls, v: str) -> StepTypeName:
        if v not in STEP_TYPES:
            raise ValueError
        return StepTypeName(v)

    @property
    def html_doc(self) -> str:
        return re.sub(r"\n\s{4}", "\n", self.doc).strip("\n ")


def public_step(func: FunctionType) -> FunctionType:
    """
    Decorator for Overhave BDD steps, enables display of decorated step.

    Attention: `OverhaveStepCollectorSettings.hide_non_public_steps` setting should be `True`!
    """
    setattr(func, PUBLIC_STEP_ATTR_NAME, True)
    return func


def is_public_step(func: FunctionType) -> bool:
    return getattr(func, PUBLIC_STEP_ATTR_NAME, False)



import logging
import re
from typing import Any, cast

import sqlalchemy as sa
import sqlalchemy.orm as so

logger = logging.getLogger(__name__)

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


class SAMetadata(sa.MetaData):
    """Custom SQLAlchemy MetaData with bounded engine."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self._engine: sa.Engine | None = None

    @property
    def engine(self) -> sa.Engine:
        if isinstance(self._engine, sa.Engine):
            return self._engine
        raise RuntimeError("MetaData has not got bounded Engine!")

    def set_engine(self, engine: sa.Engine) -> None:
        self._engine = engine


metadata = SAMetadata(naming_convention=convention)


def _classname_to_tablename(name: str) -> str:
    result: list[str] = []

    last_index = 0
    for match in re.finditer(r"(?P<abbreviation>[A-Z]+(?![a-z\d]))|(?P<word>[A-Z][a-z]*)|(?P<digit>\d+)", name):
        if match.start() != last_index:
            raise ValueError(f'Could not translate class name "{name}" to table name')

        last_index = match.end()
        result.append(match.group().lower())

    return "_".join(result)


@so.as_declarative(metadata=metadata)
class BaseTable:
    """Base table class with __tablename__."""

    @so.declared_attr.directive
    def __tablename__(cls) -> str:
        return _classname_to_tablename(cls.__name__)  # type: ignore[attr-defined]


class PrimaryKeyWithoutDateMixin:
    """Table mixin with declared attribute `id`."""

    @so.declared_attr.cascading
    def id(cls):  # type: ignore[no-untyped-def]
        return sa.Column(f"{getattr(cls, '__tablename__')}_id", sa.Integer(), primary_key=True)


class PrimaryKeyMixin(PrimaryKeyWithoutDateMixin):
    """Table mixin with `id` and `created_at` declared attributes."""

    @so.declared_attr.cascading
    def created_at(cls):  # type: ignore[no-untyped-def]
        return sa.Column(sa.DateTime(timezone=True), nullable=True, server_default=sa.func.now())


def _get_query_cls(
    mapper: tuple[type[BaseTable], ...] | so.Mapper[Any],
    session: so.Session,
) -> so.Query[Any]:
    if mapper:
        m = mapper
        if isinstance(m, tuple):
            m = mapper[0]  # type: ignore[index, assignment]
        if isinstance(m, so.Mapper):
            m = m.entity

        try:
            return cast(
                so.Query[Any],
                m.__query_cls__(mapper, session),  # type: ignore[union-attr]
            )
        except AttributeError:
            pass

    return so.Query(mapper, session)  # type: ignore[arg-type]


Session = so.sessionmaker(query_cls=_get_query_cls)

current_session = so.scoped_session(Session)



import abc
from datetime import datetime
from typing import Literal

from pydantic import BaseModel

from overhave.utils import get_current_time


class AuthToken(BaseModel):
    """Model for OAuth2 auth_managers token."""

    access_token: str
    expires_at: datetime
    token_type: Literal["Bearer"] = "Bearer"


class IAuthStorage(abc.ABC):
    """Abstract storage for auth_managers tokens."""

    @abc.abstractmethod
    def get_auth_token(self, username: str) -> AuthToken | None:
        pass

    @abc.abstractmethod
    def update_auth_token(self, username: str, new_auth_token: AuthToken) -> None:
        pass


class AuthStorage(IAuthStorage):
    """In-memory storage for auth_managers tokens."""

    def __init__(self) -> None:
        self._storage: dict[str, AuthToken] = {}

    def get_auth_token(self, username: str) -> AuthToken | None:
        auth_token = self._storage.get(username)
        if auth_token and auth_token.expires_at > get_current_time():
            return auth_token
        return None

    def update_auth_token(self, username: str, new_auth_token: AuthToken) -> None:
        self._storage[username] = new_auth_token



from datetime import datetime
from typing import cast

from jose import jwt

from overhave.api.auth.models import AuthTokenData
from overhave.api.settings import OverhaveApiAuthSettings


def create_access_token(auth_settings: OverhaveApiAuthSettings, username: str, expires_at: datetime) -> str:
    to_encode: dict[str, str | datetime] = {"sub": username, "exp": expires_at}
    return cast(
        str, jwt.encode(to_encode, auth_settings.secret_key.get_secret_value(), algorithm=auth_settings.algorithm)
    )


def get_token_data(auth_settings: OverhaveApiAuthSettings, token: str) -> AuthTokenData | None:
    payload = jwt.decode(token, auth_settings.secret_key.get_secret_value(), algorithms=[auth_settings.algorithm])
    username: str | None = payload.get("sub")
    if username is not None:
        return AuthTokenData(username=username)
    return None



import logging
from typing import Any, Generic, Mapping, cast

import httpx
import tenacity
from pydantic import BaseModel, ValidationError
from pydantic._internal._model_construction import ModelMetaclass

from overhave.transport.http.base_client.objects import HttpMethod
from overhave.transport.http.base_client.settings import HttpSettingsType

logger = logging.getLogger(__name__)


class BaseHttpClientException(Exception):
    """Base exception for BaseHttpClient."""


class HttpClientValidationError(BaseHttpClientException):
    """Model validation error for BaseHttpClient."""


class BaseHttpClient(Generic[HttpSettingsType]):
    """Base client for HTTP communications."""

    def __init__(self, settings: HttpSettingsType) -> None:
        self._settings = settings

    @staticmethod
    def _parse_or_raise(response: httpx.Response, model: ModelMetaclass) -> BaseModel:
        try:
            return cast(BaseModel, model).model_validate(response.json())
        except (ValueError, ValidationError) as e:
            url = getattr(response, "raw_url", response.url)
            raise HttpClientValidationError(f'Response validation error for "{url}"') from e

    @tenacity.retry(
        reraise=True,
        retry=tenacity.retry_if_exception_type(httpx.ConnectError),
        stop=tenacity.stop_after_attempt(3),
        before_sleep=tenacity.before_sleep_log(logger, logger.level),
        after=tenacity.after_log(logger, logger.level),
    )
    def _make_request(
        self,
        method: HttpMethod,
        url: httpx.URL,
        params: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
        data: str | bytes | Mapping[Any, Any] | None = None,
        auth: httpx.Auth | None = None,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        response = httpx.request(
            method=str(method.value),
            url=str(url),
            params=params,
            json=json,
            data=data,  # type: ignore
            auth=auth,
            timeout=self._settings.timeout,
        )
        if raise_for_status:
            response.raise_for_status()
        return response



from overhave.entities import OverhaveEmulationSettings
from overhave.factory.context.base_context import BaseFactoryContext


class OverhaveEmulationContext(BaseFactoryContext):
    """Overhave emulation context, based on application BaseSettings.

    This context defines how Overhave emulation will work.
    """

    def __init__(self, emulation_settings: OverhaveEmulationSettings | None = None) -> None:
        super().__init__(emulation_settings=emulation_settings or OverhaveEmulationSettings())



from pathlib import Path
from typing import Mapping

import httpx
from pydantic import BaseModel, field_validator, model_validator

from overhave.base_settings import BaseOverhavePrefix
from overhave.utils import make_url


class EmptyTaskTrackerURLError(ValueError):
    """Exception for situation with empty ```task_tracker_url``` while trying to ```get_task_link```."""


class EmptyGitProjectURLError(ValueError):
    """Exception for situation with empty ```git_project_url``` while trying to ```get_git_feature_url```."""


class OverhaveProjectSettings(BaseOverhavePrefix):
    """
    Entity project settings.

    You could specify your fixture content for specific pytest usage with your own plugins.
    It could give you pretty extended file with powerful possibilities.
    `fixture_context` will be formatted by `OverhaveProccessor` before test would be started.
    """

    # Fixture content in list format, which would be compiled into formatted string.
    fixture_content: list[str] = [
        "from pytest_bdd import scenarios",
        "from overhave import overhave_proxy_manager",
        "pytest_plugins = overhave_proxy_manager().plugin_resolver.get_plugins()",
        "scenarios('{feature_file_path}')",
    ]

    # Task tracker URL
    task_tracker_url: httpx.URL | None
    # Behaviour specification keyword for tasks attachment
    tasks_keyword: str | None

    # Git project URL
    git_project_url: httpx.URL | None

    # Templates are used for creation of test user and system specifications.
    # Keys for specification mapping are still the same - feature types.
    user_spec_template_mapping: Mapping[str, type[BaseModel]] = {}

    @model_validator(mode="after")
    def validate_links_keyword(self: "OverhaveProjectSettings") -> "OverhaveProjectSettings":
        if isinstance(self.tasks_keyword, str) and self.task_tracker_url is None:
            raise ValueError("'task_tracker_url' should be specified in case of 'tasks_keyword' usage!")
        return self

    @field_validator("task_tracker_url", mode="before")
    def make_tasktracker_url(cls, v: str | None) -> httpx.URL | None:
        return make_url(v)

    @field_validator("git_project_url", mode="before")
    def make_git_project_url(cls, v: str | None) -> httpx.URL | None:
        return make_url(v)

    def get_task_link(self, link: str) -> httpx.URL:
        if isinstance(self.task_tracker_url, httpx.URL):
            return httpx.URL(f"{self.task_tracker_url}/{link}")
        raise EmptyTaskTrackerURLError("Task tracker URL is None, so could not create link URL!")

    def get_git_feature_url(self, filepath: Path) -> httpx.URL:
        if isinstance(self.git_project_url, httpx.URL):
            return httpx.URL(f"{self.git_project_url}/{filepath.as_posix()}")
        raise EmptyGitProjectURLError("Git project URL is None, so could not create link URL!")



import abc
from datetime import datetime

import sqlalchemy as sa
import sqlalchemy.orm as so

from overhave import db
from overhave.storage import DraftModel
from overhave.utils import get_current_time


class IDraftStorage(abc.ABC):
    """Abstract class for scenario versions storage."""

    @staticmethod
    @abc.abstractmethod
    def draft_model_by_id(session: so.Session, draft_id: int) -> DraftModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_last_published_at_for_feature(feature_id: int) -> datetime | None:
        pass

    @classmethod
    @abc.abstractmethod
    def get_or_create_draft(
        cls, session: so.Session, test_run: db.TestRun, published_by: str, status: db.DraftStatus
    ) -> int:
        pass

    @staticmethod
    @abc.abstractmethod
    def save_response_as_created(draft_id: int, pr_url: str, published_at: datetime) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def save_response_as_duplicate(draft_id: int, feature_id: int, traceback: str | None) -> None:
        pass

    @abc.abstractmethod
    def set_draft_status(self, draft_id: int, status: db.DraftStatus, traceback: str | None = None) -> None:
        pass


class BaseDraftStorageException(Exception):
    """Base exception for :class:`DraftStorage`."""


class NullableScenarioError(BaseDraftStorageException):
    """Exception for situation when Scenario text is None."""


class DraftNotFoundError(BaseDraftStorageException):
    """Exception for situation when draft not found."""


class FeatureNotFoundError(BaseDraftStorageException):
    """Exception for situation when feature not found by feature_id from draft."""


class DraftStorage(IDraftStorage):
    """Class for scenario versions storage."""

    @staticmethod
    def draft_model_by_id(session: so.Session, draft_id: int) -> DraftModel:
        draft = session.query(db.Draft).filter(db.Draft.id == draft_id).one()
        return DraftModel.model_validate(draft)

    @staticmethod
    def get_last_published_at_for_feature(feature_id: int) -> datetime | None:
        with db.create_session() as session:
            last_draft_with_published_at = (
                session.query(db.Draft)
                .filter(db.Draft.feature_id == feature_id, db.Draft.published_at.isnot(None))
                .order_by(db.Draft.id.desc())
                .first()
            )
            if last_draft_with_published_at is None:
                return None
            return last_draft_with_published_at.published_at

    @staticmethod
    def _get_draft_id_by_testrun_id(session: so.Session, test_run_id: int) -> int | None:
        draft = session.query(db.Draft).filter(db.Draft.test_run_id == test_run_id).one_or_none()
        if draft is None:
            return None
        return draft.id

    @staticmethod
    def _create_draft(session: so.Session, test_run: db.TestRun, published_by: str, status: db.DraftStatus) -> int:
        text = test_run.scenario.text
        if text is None:
            raise NullableScenarioError(f"TestRun.id={test_run.id} has Scenario without text!")
        draft = db.Draft(
            feature_id=test_run.scenario.feature_id,
            test_run_id=test_run.id,
            text=text,
            published_by=published_by,
            status=status,
        )
        session.add(draft)
        session.flush()
        return draft.id

    @classmethod
    def get_or_create_draft(
        cls, session: so.Session, test_run: db.TestRun, published_by: str, status: db.DraftStatus
    ) -> int:
        draft_id = cls._get_draft_id_by_testrun_id(session=session, test_run_id=test_run.id)
        if draft_id is not None:
            return draft_id
        return cls._create_draft(session=session, test_run=test_run, published_by=published_by, status=status)

    @staticmethod
    def save_response_as_created(draft_id: int, pr_url: str, published_at: datetime) -> None:
        with db.create_session() as session:
            session.execute(
                sa.update(db.Draft)
                .where(db.Draft.id == draft_id)
                .values(
                    pr_url=pr_url,
                    published_at=published_at,
                    status=db.DraftStatus.CREATED,
                )
            )
            feature_id_query = (
                session.query(db.Draft)
                .with_entities(db.Draft.feature_id)
                .filter(db.Draft.id == draft_id)
                .scalar_subquery()
            )
            session.execute(
                sa.update(db.Feature).where(db.Feature.id == feature_id_query).values(released=True),
                execution_options={"synchronize_session": "fetch"},
            )

    @staticmethod
    def save_response_as_duplicate(draft_id: int, feature_id: int, traceback: str | None) -> None:
        with db.create_session() as session:
            last_drafts = (
                session.query(db.Draft)
                .filter(db.Draft.feature_id == feature_id)
                .order_by(db.Draft.id.desc())
                .limit(2)
                .all()
            )
            if not last_drafts:
                raise DraftNotFoundError("No one draft exists for Feature.id=%s", feature_id)
            values = dict(status=db.DraftStatus.DUPLICATE, traceback=traceback, published_at=get_current_time())

            # check that previous draft exists and succeed
            if len(last_drafts) > 1 and last_drafts[1].status.is_succeed:
                previous_draft = last_drafts[1]
                values["pr_url"] = previous_draft.pr_url
                values["published_at"] = previous_draft.published_at
            session.execute(sa.update(db.Draft).where(db.Draft.id == draft_id).values(**values))

    def set_draft_status(self, draft_id: int, status: db.DraftStatus, traceback: str | None = None) -> None:
        with db.create_session() as session:
            values: dict[str, db.DraftStatus | str | None] = dict(status=status)
            if isinstance(traceback, str):
                values["traceback"] = traceback
            session.execute(sa.update(db.Draft).where(db.Draft.id == draft_id).values(**values))



import logging
from functools import cache

from redis import Redis
from redis.sentinel import Sentinel

from overhave.transport.redis.settings import BaseRedisSettings, OverhaveRedisSentinelSettings, OverhaveRedisSettings

logger = logging.getLogger(__name__)


def make_sentinel_master(settings: OverhaveRedisSentinelSettings) -> Redis:  # type: ignore
    logger.info("Connecting to redis through sentinel %s", settings.urls)
    url_tuples = [(url.host, url.port) for url in settings.urls if url.host is not None and url.port is not None]
    sentinel = Sentinel(url_tuples, socket_timeout=settings.socket_timeout.total_seconds(), retry_on_timeout=True)
    return sentinel.master_for(settings.master_set, password=settings.password, db=settings.db)


def make_regular_redis(redis_settings: OverhaveRedisSettings) -> Redis:  # type: ignore
    return Redis.from_url(
        redis_settings.url.human_repr(),
        db=redis_settings.db,
        socket_timeout=redis_settings.socket_timeout.total_seconds(),
    )


def make_redis(redis_settings: BaseRedisSettings) -> Redis:  # type: ignore  # noqa: E501
    if isinstance(redis_settings, OverhaveRedisSentinelSettings):
        return make_sentinel_master(redis_settings)

    if isinstance(redis_settings, OverhaveRedisSettings):
        return make_regular_redis(redis_settings)

    raise RuntimeError("redis_settings is not any instance of OverhaveRedisSentinelSettings, OverhaveRedisSettings")


@cache
def get_redis_settings() -> BaseRedisSettings:
    sentinel_settings = OverhaveRedisSentinelSettings()
    if sentinel_settings.enabled:
        return sentinel_settings
    return OverhaveRedisSettings()



class BaseFeatureValidatorException(Exception):
    """Base exception for :class:`FeatureValidator`."""


class FeaturesWithoutIDPresenceError(BaseFeatureValidatorException):
    """Exception for situation with nullable ID features presence."""


class IncorrectFeaturesPresenceError(BaseFeatureValidatorException):
    """Exception for situation with incorrect features presence."""


class DuplicateFeatureIDError(BaseFeatureValidatorException):
    """Exception for situation with duplicate feature IDs."""



import logging
import re
from dataclasses import asdict
from datetime import datetime
from functools import cached_property
from typing import Sequence

import allure
from pytest_bdd import types as default_types

from overhave.entities import IFeatureExtractor, OverhaveLanguageSettings
from overhave.scenario.compiler import OverhaveScenarioCompilerSettings
from overhave.scenario.parser.models import FeatureInfo, StrictFeatureInfo
from overhave.scenario.parser.settings import OverhaveScenarioParserSettings
from overhave.scenario.prefix_mixin import PrefixMixin
from overhave.storage import FeatureTypeName

logger = logging.getLogger(__name__)
_DEFAULT_ID = 1


class BaseScenarioParserError(Exception):
    """Base exception for parsing error."""


class FeatureNameParsingError(BaseScenarioParserError):
    """Exception for feature name parsing error."""


class FeatureTypeParsingError(BaseScenarioParserError):
    """Exception for feature type parsing error."""


class AdditionalInfoParsingError(BaseScenarioParserError):
    """Exception for additional info parsing error."""


class DatetimeParsingError(BaseScenarioParserError):
    """Exception for datetime parsing error."""


class StrictFeatureParsingError(BaseScenarioParserError):
    """Exception for ValueError while initialization of StrictFeatureInfo."""


class NullableFeatureIdError(BaseScenarioParserError):
    """Exception for situation when feature id not specified (it's manually created feature)."""


class ScenarioParser(PrefixMixin):
    """Class for feature files parsing."""

    def __init__(
        self,
        parser_settings: OverhaveScenarioParserSettings,
        compilation_settings: OverhaveScenarioCompilerSettings,
        language_settings: OverhaveLanguageSettings,
        feature_extractor: IFeatureExtractor,
        tasks_keyword: str | None,
    ) -> None:
        self._parser_settings = parser_settings
        self._compilation_settings = compilation_settings
        self._language_settings = language_settings
        self._feature_extractor = feature_extractor
        self._tasks_keyword = tasks_keyword

    def set_strict_mode(self, mode: bool) -> None:
        self._parser_settings.parser_strict_mode = mode

    @cached_property
    def _feature_prefixes(self) -> list[str]:
        prefixes = [self._as_prefix(default_types.FEATURE)]
        if self._language_settings.step_prefixes is not None:
            prefixes.append(self._as_prefix(self._language_settings.step_prefixes.FEATURE))
        logger.debug("Cached ScenarioParser feature prefixes: %s", prefixes)
        return prefixes

    @cached_property
    def _task_prefix(self) -> str | None:
        if isinstance(self._tasks_keyword, str):
            return self._as_prefix(self._tasks_keyword)
        return None

    def _get_id(self, id_line: str) -> int:
        return int(id_line.removeprefix(self._compilation_settings.id_prefix).strip())

    def _get_name(self, name_line: str, feature_prefix: str) -> str:
        name_parts = name_line.split(feature_prefix)
        if not name_parts:
            raise FeatureNameParsingError(
                f"Could not parse feature name from '{name_line}'!",
            )
        return name_parts[-1].strip()

    def _get_tags(self, tags_line: str) -> list[str]:
        return [tag.strip() for tag in tags_line.split(self._compilation_settings.tag_prefix) if tag]

    def _get_feature_type(self, tags: Sequence[str]) -> FeatureTypeName:
        for tag in tags:
            if tag not in self._feature_extractor.feature_types:
                continue
            return FeatureTypeName(tag)
        raise FeatureTypeParsingError(
            f"Could not get feature type from tags {tags}!",
        )

    def _get_severity_tag(self, tags: Sequence[str]) -> str | None:
        for tag in reversed(tags):
            if self._compilation_settings.severity_keyword not in tag:
                continue
            return tag
        return None

    @staticmethod
    def _get_additional_info(additional_line: str, left_pointer: str, right_pointer: str) -> str:
        result = re.search(rf"({left_pointer})+\s?[\w.]+\s?({right_pointer})+", additional_line)
        if result:
            return result.group(0).removeprefix(left_pointer).removesuffix(right_pointer).strip()
        raise AdditionalInfoParsingError("Could not parse additional info from '%s'!", additional_line)

    def _get_task_info(self, task_line: str) -> list[str]:
        tasks = []
        if self._task_prefix is not None:
            tasks.extend(task_line.removeprefix(self._task_prefix).split(","))
        return [x.strip() for x in tasks]

    def _get_time_info(self, line: str, left_pointer: str, right_pointer: str) -> datetime:
        result = re.search(rf"({left_pointer})+[\w\-:\s]+({right_pointer})+", line)
        if result:
            datetime_str = result.group(0).removeprefix(left_pointer).removesuffix(right_pointer).strip()
            return datetime.strptime(datetime_str, self._compilation_settings.time_format)
        raise DatetimeParsingError("Could not parse datetime from '%s'!", line)

    def _parse_feature_info(self, header: str) -> FeatureInfo:  # noqa: C901
        feature_info = FeatureInfo()
        for line in header.split("\n"):
            if line.startswith(self._compilation_settings.id_prefix):
                feature_info.id = self._get_id(line)
                continue
            if line.startswith(self._compilation_settings.tag_prefix):
                tags = self._get_tags(line)
                feature_info.type = self._get_feature_type(tags)
                tags.remove(feature_info.type)

                severity_tag = self._get_severity_tag(tags)
                if severity_tag is not None:
                    tags.remove(severity_tag)
                    feature_info.severity = allure.severity_level(
                        severity_tag.removeprefix(self._compilation_settings.severity_keyword)
                    )

                feature_info.tags = tags
                continue
            for prefix in self._feature_prefixes:
                if not line.startswith(prefix):
                    continue
                feature_info.name = self._get_name(name_line=line, feature_prefix=prefix)
            if self._compilation_settings.created_by_prefix in line:
                feature_info.author = self._get_additional_info(
                    line,
                    left_pointer=self._compilation_settings.created_by_prefix,
                    right_pointer=self._compilation_settings.blocks_delimiter,
                )
            if self._compilation_settings.last_edited_by_prefix in line:
                feature_info.last_edited_by = self._get_additional_info(
                    line,
                    left_pointer=self._compilation_settings.last_edited_by_prefix,
                    right_pointer=self._compilation_settings.time_delimiter,
                )
                feature_info.last_edited_at = self._get_time_info(
                    line,
                    left_pointer=self._compilation_settings.time_delimiter,
                    right_pointer=self._compilation_settings.blocks_delimiter,
                )
            if self._task_prefix is not None and self._task_prefix in line:
                feature_info.tasks = self._get_task_info(line)
                continue
        if feature_info.name is None:
            raise FeatureNameParsingError(f"Could not parse feature name from header:\n{header}")
        return feature_info

    def parse(self, feature_txt: str) -> FeatureInfo | StrictFeatureInfo:
        blocks_delimiter = "\n\n"
        indent = "  "
        indent_substitute = "__"
        blocks = [
            block.lstrip(" \n") for block in feature_txt.replace(indent, indent_substitute).split(blocks_delimiter)
        ]
        header = blocks.pop(0)
        feature_info = self._parse_feature_info(header)
        feature_info.scenarios = blocks_delimiter.join(blocks).replace(indent_substitute, indent)
        if not self._parser_settings.parser_strict_mode:
            return feature_info
        if feature_info.id is None:
            raise NullableFeatureIdError("Feature has not got specified ID!")
        try:
            return StrictFeatureInfo(**asdict(feature_info))
        except ValueError as err:
            raise StrictFeatureParsingError("Could not parse feature to StrictFeatureInfo!") from err



from datetime import timedelta
from typing import TypeVar

import httpx
from pydantic import field_validator
from pydantic_settings import BaseSettings

from overhave.utils import make_url


class BaseHttpClientSettings(BaseSettings):
    """Base settings for :class:`BaseHttpClient`."""

    url: httpx.URL

    default_timeout: timedelta | None = None
    connect_timeout: timedelta = timedelta(seconds=1)
    read_timeout: timedelta = timedelta(seconds=5)
    write_timeout: timedelta | None = None
    pool_timeout: timedelta | None = None

    @field_validator("url", mode="before")
    def make_url(cls, v: str | None) -> httpx.URL | None:
        return make_url(v)

    @staticmethod
    def _as_optional_timeout(timeout: timedelta | None) -> float | None:
        if timeout is not None:
            return timeout.total_seconds()
        return None

    @property
    def timeout(self) -> httpx.Timeout:
        return httpx.Timeout(
            self._as_optional_timeout(self.default_timeout),
            connect=self.connect_timeout.total_seconds(),
            read=self.read_timeout.total_seconds(),
            write=self._as_optional_timeout(self.write_timeout),
            pool=self._as_optional_timeout(self.pool_timeout),
        )


HttpSettingsType = TypeVar("HttpSettingsType", bound=BaseHttpClientSettings)



from typing import Literal

from pydantic.main import BaseModel


class AuthTokenData(BaseModel):
    """Model for OAuth2 auth_managers token data."""

    username: str


class AuthHeaders(BaseModel):
    """Model for OAuth2 auth_managers HTTP headers."""

    Authorization: Literal["Bearer"] = "Bearer"


AUTH_HEADERS: AuthHeaders = AuthHeaders()



from __future__ import annotations

import datetime
from typing import Dict, List

import allure
import sqlalchemy as sa
import sqlalchemy_utils as su
from flask import url_for
from sqlalchemy import orm as so

from overhave.db.base import BaseTable, PrimaryKeyMixin, PrimaryKeyWithoutDateMixin
from overhave.db.statuses import DraftStatus, EmulationStatus, TestReportStatus, TestRunStatus
from overhave.db.users import UserRole


class FeatureType(BaseTable, PrimaryKeyWithoutDateMixin):
    """Feature types table."""

    name: str = sa.Column(sa.Text, unique=True, nullable=False, doc="Feature types choice")

    def __repr__(self) -> str:
        return self.name.upper()


class Tags(BaseTable, PrimaryKeyMixin):
    """Feature tags table."""

    value: str = sa.Column(sa.String(), nullable=False, unique=True, doc="Feature tags choice")
    created_by: str = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), nullable=False, doc="Author login")

    def __repr__(self) -> str:
        return self.value


@su.generic_repr("id", "name", "last_edited_by")
class Feature(BaseTable, PrimaryKeyMixin):
    """Features table."""

    name: str = sa.Column(sa.String(), doc="Feature name", nullable=False, unique=True)
    author: str = sa.Column(
        sa.String(), sa.ForeignKey(UserRole.login), doc="Feature author login", nullable=False, index=True
    )
    type_id: int = sa.Column(sa.Integer(), sa.ForeignKey(FeatureType.id), nullable=False, doc="Feature types choice")
    file_path: str = sa.Column(sa.String(), doc="Feature file path", nullable=False, unique=True)
    task: List[str] = sa.Column(sa.ARRAY(sa.String()), doc="Feature tasks list", nullable=False)
    last_edited_by: str = sa.Column(sa.String(), doc="Last feature editor login", nullable=False)
    last_edited_at: datetime.datetime = sa.Column(
        sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()
    )
    released: bool = sa.Column(sa.Boolean, doc="Feature release state boolean", nullable=False, default=False)
    severity: allure.severity_level = sa.Column(
        sa.Enum(allure.severity_level),
        nullable=False,
        default=allure.severity_level.NORMAL,
        doc="Feature severity choice",
    )

    feature_type: so.Mapped[FeatureType] = so.relationship(FeatureType, uselist=False)
    feature_tags: so.Mapped[List[Tags]] = so.relationship(
        Tags, uselist=True, order_by=Tags.value, secondary="feature_tags_association_table"
    )


class FeatureTagsAssociationTable(BaseTable, PrimaryKeyWithoutDateMixin):
    """Association table between features and tags."""

    extend_existing = True

    tags_id: int = sa.Column(sa.Integer(), sa.ForeignKey(Tags.id))
    feature_id: int = sa.Column(sa.Integer(), sa.ForeignKey(Feature.id))


@su.generic_repr("feature_id")
class Scenario(BaseTable, PrimaryKeyMixin):
    """Scenarios table."""

    feature_id: int = sa.Column(sa.Integer(), sa.ForeignKey(Feature.id), nullable=False, unique=True)
    text: str | None = sa.Column(sa.Text(), doc="Text storage for scenarios in feature")

    feature: so.Mapped[Feature] = so.relationship(
        Feature, uselist=False, backref=so.backref("scenario", cascade="all, delete-orphan")
    )


class TestRun(BaseTable, PrimaryKeyMixin):
    """Test runs table."""

    __test__ = False

    scenario_id: int = sa.Column(sa.Integer(), sa.ForeignKey(Scenario.id), nullable=False, index=True)
    name: str = sa.Column(sa.String(), nullable=False)
    start: datetime.datetime | None = sa.Column(sa.DateTime(timezone=True), doc="Test start time")
    end: datetime.datetime | None = sa.Column(sa.DateTime(timezone=True), doc="Test finish time")
    status: TestRunStatus = sa.Column(sa.Enum(TestRunStatus), doc="Current test status", nullable=False)
    report_status: TestReportStatus = sa.Column(
        sa.Enum(TestReportStatus), doc="Report generation result", nullable=False
    )
    executed_by: str = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Test executor login", nullable=False)
    report: str | None = sa.Column(sa.String(), doc="Relative report URL")
    traceback: str | None = sa.Column(sa.Text(), doc="Text storage for error traceback")

    scenario: so.Mapped[Scenario] = so.relationship(
        Scenario, uselist=False, backref=so.backref("test_runs", cascade="all, delete-orphan")
    )


class Draft(BaseTable, PrimaryKeyMixin):
    """Scenario versions table."""

    feature_id: int = sa.Column(sa.Integer(), sa.ForeignKey(Feature.id), nullable=False, index=True)
    test_run_id: int = sa.Column(sa.Integer(), sa.ForeignKey(TestRun.id), unique=True, nullable=False)
    text: str | None = sa.Column(sa.Text(), doc="Released scenario text")
    pr_url: str | None = sa.Column(sa.String(), doc="Absolute pull-request URL")
    published_by: str = sa.Column(
        sa.String(), sa.ForeignKey(UserRole.login), doc="Draft publisher login", nullable=False
    )
    published_at: datetime.datetime | None = sa.Column(sa.DateTime(timezone=True), doc="Publication time")
    traceback: str | None = sa.Column(sa.Text(), doc="Text storage for error traceback")
    status: DraftStatus = sa.Column(sa.Enum(DraftStatus), doc="Version publishing status", nullable=False)

    feature: so.Mapped[Feature] = so.relationship(
        Feature, uselist=False, backref=so.backref("versions", cascade="all, delete-orphan")
    )

    def __html__(self) -> str:
        return f'<a href="{url_for("draft.details_view", id=self.id)}">Draft: {self.id}</a>'


@su.generic_repr("id", "name", "created_by")
class TestUser(BaseTable, PrimaryKeyMixin):
    """Test users table."""

    __test__ = False

    key: str = sa.Column(sa.String(), nullable=False, unique=True, doc="Unique user key")
    name: str = sa.Column(sa.String(), nullable=False, unique=True, doc="Informational user name")
    feature_type_id: int = sa.Column(
        sa.Integer(), sa.ForeignKey(FeatureType.id), nullable=False, doc="Feature types choice"
    )
    specification: Dict[str, str | None] = sa.Column(sa.JSON(none_as_null=True))
    created_by: str = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)
    allow_update: bool = sa.Column(sa.Boolean(), doc="User updating allowance", nullable=False, default=False)
    changed_at: datetime.datetime = sa.Column(sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now())

    feature_type: so.Mapped[FeatureType] = so.relationship(FeatureType, uselist=False)


class Emulation(BaseTable, PrimaryKeyMixin):
    """Emulation templates table."""

    name: str = sa.Column(sa.String(), nullable=False, unique=True)
    test_user_id: int = sa.Column(sa.Integer(), sa.ForeignKey(TestUser.id), nullable=False, doc="Test user ID")
    command: str = sa.Column(sa.String(), nullable=False, doc="Command for emulator's execution")
    created_by: str = sa.Column(sa.String(), sa.ForeignKey(UserRole.login), doc="Author login", nullable=False)

    test_user: so.Mapped[TestUser] = so.relationship(
        TestUser, uselist=False, backref=so.backref("emulations", cascade="all, delete-orphan")
    )


class EmulationRun(BaseTable, PrimaryKeyMixin):
    """Emulation runs table."""

    emulation_id: int = sa.Column(sa.Integer(), sa.ForeignKey(Emulation.id), nullable=False, index=True)
    status: EmulationStatus = sa.Column(sa.Enum(EmulationStatus), doc="Current emulation status", nullable=False)
    changed_at: datetime.datetime = sa.Column(sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now())
    port: int | None = sa.Column(sa.Integer(), doc="Port for emulation")
    traceback: str | None = sa.Column(sa.Text(), doc="Text storage for error traceback")
    initiated_by: str = sa.Column(
        sa.String(), sa.ForeignKey(UserRole.login), doc="Initiator of start emulation", nullable=False
    )

    emulation: so.Mapped[Emulation] = so.relationship(
        Emulation, uselist=False, backref=so.backref("emulation_runs", cascade="all, delete-orphan")
    )



from overhave.api.app import create_overhave_api
from overhave.base_settings import DataBaseSettings, LoggingSettings

DataBaseSettings().setup_engine()
LoggingSettings().setup_logging()

app = create_overhave_api()



from datetime import timedelta

import yarl
from pydantic import field_validator
from pydantic_settings import BaseSettings


class BaseRedisSettings(BaseSettings):
    """Base settings for Redis entities, which use for work with different framework tasks."""

    db: int = 0
    block_timeout: timedelta = timedelta(seconds=1)
    read_count: int = 1
    socket_timeout: timedelta = timedelta(seconds=5)

    @property
    def timeout_milliseconds(self) -> int:
        return int(self.block_timeout.total_seconds() * 1000)

    class Config:
        env_prefix = "OVERHAVE_REDIS_"


class OverhaveRedisSettings(BaseRedisSettings):
    """Settings for Redis entities, which use for work with different framework tasks."""

    url: yarl.URL = yarl.URL("redis://localhost:6379")

    @field_validator("url", mode="before")
    def validate_url(cls, v: str | yarl.URL) -> yarl.URL:
        if isinstance(v, str):
            return yarl.URL(v)
        return v


class OverhaveRedisSentinelSettings(BaseRedisSettings):
    """Settings for Redis sentinel entities, which use for work with different framework tasks."""

    enabled: bool = False
    urls: list[yarl.URL] = [yarl.URL("redis://localhost:6379")]
    master_set: str = "foo"
    password: str = "bar"

    @field_validator("urls", mode="before")
    def validate_urls(cls, v: list[str] | list[yarl.URL]) -> list[yarl.URL]:
        urls = []
        for url in v:
            if isinstance(url, str):
                urls.append(yarl.URL(url))
            if isinstance(url, yarl.URL):
                urls.append(url)
        return urls

    class Config:
        env_prefix = "OVERHAVE_REDIS_SENTINEL_"



import enum


class TestRunStatus(enum.StrEnum):
    """Enum for test run statuses."""

    __test__ = False

    STARTED = "STARTED"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"
    INTERNAL_ERROR = "INTERNAL_ERROR"

    @property
    def finished(self) -> bool:
        return self in (TestRunStatus.FAILED, TestRunStatus.SUCCESS, TestRunStatus.INTERNAL_ERROR)


class TestReportStatus(str, enum.Enum):
    """Enum for test run statuses."""

    __test__ = False

    EMPTY = "EMPTY"
    GENERATION_FAILED = "GENERATION_FAILED"
    GENERATED = "GENERATED"
    SAVED = "SAVED"

    @property
    def has_report(self) -> bool:
        return self in (TestReportStatus.GENERATED, TestReportStatus.SAVED)


class EmulationStatus(enum.StrEnum):
    """Enum for emulation statuses."""

    CREATED = "CREATED"
    REQUESTED = "REQUESTED"
    READY = "READY"
    ERROR = "ERROR"

    @property
    def processed(self) -> bool:
        return self in (EmulationStatus.READY, EmulationStatus.ERROR)


class DraftStatus(enum.StrEnum):
    """Enum for draft statuses."""

    REQUESTED = "REQUESTED"
    CREATING = "CREATING"
    CREATED = "CREATED"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    DUPLICATE = "DUPLICATE"

    @property
    def is_succeed(self) -> bool:
        return self in (DraftStatus.CREATED, DraftStatus.DUPLICATE)



# flake8: noqa
from .consumer import RedisConsumer
from .objects import (
    AnyRedisTask,
    BaseRedisTask,
    EmulationData,
    EmulationTask,
    PublicationData,
    PublicationTask,
    RedisStream,
    TestRunData,
    TestRunTask,
    TRedisTask,
)
from .producer import RedisProducer
from .runner import RedisConsumerRunner
from .settings import BaseRedisSettings, OverhaveRedisSentinelSettings, OverhaveRedisSettings



from functools import cache

from prometheus_client import CollectorRegistry

from overhave.metrics import (
    BaseOverhaveMetricContainer,
    EmulationRunOverhaveMetricContainer,
    PublicationOverhaveMetricContainer,
    TestRunOverhaveMetricContainer,
)


def get_collector_registry() -> CollectorRegistry:
    return CollectorRegistry()


@cache
def get_common_metric_container() -> BaseOverhaveMetricContainer:
    return BaseOverhaveMetricContainer(registry=get_collector_registry())


@cache
def get_test_metric_container() -> TestRunOverhaveMetricContainer:
    return TestRunOverhaveMetricContainer(registry=get_collector_registry())


@cache
def get_emulation_metric_container() -> EmulationRunOverhaveMetricContainer:
    return EmulationRunOverhaveMetricContainer(registry=get_collector_registry())


@cache
def get_publication_metric_container() -> PublicationOverhaveMetricContainer:
    return PublicationOverhaveMetricContainer(registry=get_collector_registry())



from typing import Literal

from pydantic import BaseModel


class TokenRequestData(BaseModel):
    """Model for OAuth2 request data."""

    grant_type: Literal["password"] = "password"
    username: str
    password: str



import logging
from pathlib import Path
from typing import Any

import flask
import httpx
import ldap
from flask_admin import AdminIndexView, expose
from flask_login import login_required, login_user, logout_user
from markupsafe import Markup
from sqlalchemy.exc import OperationalError, ProgrammingError
from werkzeug.wrappers import Response
from wtforms.validators import ValidationError

from overhave.admin.views.index.custom_page import CustomPageForm
from overhave.admin.views.index.login_form import LoginForm
from overhave.entities import IAdminAuthorizationManager

logger = logging.getLogger(__name__)


class OverhaveIndexView(AdminIndexView):
    """View for index."""

    def __init__(
        self,
        name: str,
        url: str,
        auth_manager: IAdminAuthorizationManager,
        index_template_path: Path | None,
        support_chat_url: httpx.URL | None,
    ) -> None:
        super().__init__(
            name=name,
            url=url,
        )
        self._auth_manager = auth_manager
        self._index_template_path = index_template_path
        self._support_chat_url = support_chat_url

    @expose("/login", methods=["GET", "POST"])
    def login(self) -> Any:  # noqa: C901
        form = LoginForm(auth_manager=self._auth_manager)
        if not form.validate_on_submit():
            return self.render("login.html", form=form)
        try:
            user = form.get_user()
            login_user(user)
        except ValidationError:
            flash_msg = f"Username '{form.username.data}' is not registered!"
            if self._support_chat_url:
                flash_msg += f" Please contact the <a href='{self._support_chat_url}'>support channel</a>!"

            return form.flash_and_redirect(Markup(flash_msg))
        except ldap.SERVER_DOWN:
            return form.flash_and_redirect("LDAP auth_managers service is unreachable!")
        except OperationalError:
            logger.exception("DataBase is unreachable!")
            return form.flash_and_redirect("DataBase is unreachable!")
        except ProgrammingError:
            logger.exception("Error while trying to operate with DataBase")
            return form.flash_and_redirect("Error while trying to operate with DataBase")
        except Exception:
            logger.exception("Unknown error!")
            return form.flash_and_redirect("Unknown error!")
        next_path = flask.request.args.get("next")
        if next_path:
            logger.info("Redirect to %s...", next_path)
            return flask.redirect(next_path)
        return flask.redirect(flask.url_for("admin.index"))

    @expose("/logout", methods=["GET"])
    @login_required
    def logout(self) -> Response:
        logout_user()
        return flask.redirect(flask.url_for("admin.login"))

    @expose("/", methods=["GET", "POST"])  # noqa: C901
    @login_required
    def index(self) -> Any:  # noqa: C901
        return self.render("index.html", form=CustomPageForm(template_path=self._index_template_path))



import typer as typer

from overhave.cli.db_cmds import db_app
from overhave.cli.s3 import s3_app
from overhave.cli.synchronizer import sync_app

overhave: typer.Typer = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})
overhave.add_typer(db_app, name="db")
overhave.add_typer(s3_app, name="s3")
overhave.add_typer(sync_app, name="sync")



import logging
from pprint import pformat

import sqlalchemy.orm as so
from pydantic import SecretStr

from overhave import db
from overhave.entities.auth_managers.base import BaseAdminAuthorizationManager
from overhave.entities.auth_managers.ldap.settings import OverhaveLdapManagerSettings
from overhave.storage import ISystemUserGroupStorage, ISystemUserStorage, SystemUserModel
from overhave.transport.ldap.authenticator import LDAPAuthenticator

logger = logging.getLogger(__name__)


class LDAPEmptyAdminGroupError(ValueError):
    """Exception for situation with empty LDAP admin group."""


class LDAPAdminAuthorizationManager(BaseAdminAuthorizationManager):
    """Class for user auth_managers via LDAP.

    Manager authorize users using remote LDAP server. Each user should use his LDAP credentials.
    LDAP server returns user groups. If user in default 'admin' group or his groups list contains admin group - user
    will be authorized. If user already placed in database - user will be authorized too. No one password stores.
    """

    def __init__(
        self,
        settings: OverhaveLdapManagerSettings,
        system_user_storage: ISystemUserStorage,
        system_user_group_storage: ISystemUserGroupStorage,
        ldap_authenticator: LDAPAuthenticator,
    ):
        super().__init__(system_user_storage)
        self._settings = settings
        self._system_user_group_storage = system_user_group_storage
        self._ldap_authenticator = ldap_authenticator

    @staticmethod
    def _user_role_by_admin_group_entry(user_has_admin_group: bool) -> db.Role:
        if user_has_admin_group:
            return db.Role.admin
        return db.Role.user

    def _can_user_be_created(self, session: so.Session, user_has_admin_group: bool, user_groups: list[str]) -> bool:
        return user_has_admin_group or self._system_user_group_storage.has_any_group(
            session=session, user_groups=user_groups
        )

    def authorize_user(self, username: str, password: SecretStr) -> SystemUserModel | None:
        logger.debug("Try to authorize user '%s'...", username)
        user_groups = self._ldap_authenticator.get_user_groups(username, password)
        if not user_groups:
            logger.info("LDAP user '%s' does not exist!", username)
            return None
        logger.debug("LDAP user groups: \n %s", pformat(user_groups))

        with db.create_session() as session:
            user_has_admin_group = self._settings.ldap_admin_group in user_groups
            intended_user_role = self._user_role_by_admin_group_entry(user_has_admin_group)

            user = self._system_user_storage.get_user_by_credits(session=session, login=username)
            if user is not None:
                if user.role is db.Role.user and user.role is not intended_user_role:
                    user.role = intended_user_role
                    session.flush()
                return SystemUserModel.model_validate(user)

            logger.debug("Have not found user with username '%s'!", username)
            if self._can_user_be_created(
                session=session, user_has_admin_group=user_has_admin_group, user_groups=user_groups
            ):
                user = self._system_user_storage.create_user(session=session, login=username, role=intended_user_role)
                return SystemUserModel.model_validate(user)

        logger.debug("Received user groups (%s) are not supplied with exist groups!", user_groups)
        return None



# flake8: noqa
from .client import GitlabHttpClient
from .models import GitlabMrCreationResponse, GitlabMrRequest, GitlabRepository
from .objects import TokenType
from .settings import OverhaveGitlabClientSettings
from .utils import get_gitlab_python_client



from typing import Any

from pydantic import Field, model_validator
from pydantic_settings import BaseSettings


class OverhaveS3ManagerSettings(BaseSettings):
    """Settings for S3Client."""

    enabled: bool = False

    url: str | None = Field(default=None)
    region_name: str | None = Field(default=None)
    access_key: str | None = Field(default=None)
    secret_key: str | None = Field(default=None)
    verify: bool = True

    autocreate_buckets: bool = False

    class Config:
        env_prefix = "OVERHAVE_S3_"

    @model_validator(mode="before")
    def validate_enabling(cls, values: dict[str, Any]) -> dict[str, Any]:
        enabled = values.get("enabled")
        if enabled and not all(
            (
                isinstance(values.get("url"), str),
                isinstance(values.get("access_key"), str),
                isinstance(values.get("secret_key"), str),
            )
        ):
            raise ValueError("Url, access key and secret key should be specified!")
        return values



from http import HTTPStatus

import fastapi

from overhave.api.deps import (
    get_feature_storage,
    get_feature_tag_storage,
    get_redis_producer,
    get_scenario_storage,
    get_test_run_storage,
)
from overhave.api.views.tags_views import tags_item_handler
from overhave.storage import IFeatureStorage, IFeatureTagStorage, IScenarioStorage, TestRunModel, TestRunStorage
from overhave.transport import RedisProducer, TestRunData, TestRunTask


def get_test_run_handler(
    test_run_id: int,
    test_run_storage: TestRunStorage = fastapi.Depends(get_test_run_storage),
) -> TestRunModel:
    test_run = test_run_storage.get_testrun_model(test_run_id)
    if test_run is not None:
        return test_run
    raise fastapi.HTTPException(
        status_code=HTTPStatus.BAD_REQUEST, detail=f"Test run with id ='{test_run_id}' not found"
    )


def run_tests_by_tag_handler(
    tag_value: str,
    feature_storage: IFeatureStorage = fastapi.Depends(get_feature_storage),
    tag_storage: IFeatureTagStorage = fastapi.Depends(get_feature_tag_storage),
    scenario_storage: IScenarioStorage = fastapi.Depends(get_scenario_storage),
    test_run_storage: TestRunStorage = fastapi.Depends(get_test_run_storage),
    redis_producer: RedisProducer = fastapi.Depends(get_redis_producer),
) -> list[str]:
    tag_model = tags_item_handler(value=tag_value, feature_tag_storage=tag_storage)
    features = feature_storage.get_features_by_tag(tag_id=tag_model.id)
    if not features:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"Features with tag='{tag_value}' do not exist"
        )
    test_run_ids: list[str] = []
    for feature in features:
        scenario = scenario_storage.get_scenario_by_feature_id(feature.id)
        test_run_id = test_run_storage.create_testrun(scenario_id=scenario.id, executed_by=feature.last_edited_by)
        redis_producer.add_task(TestRunTask(data=TestRunData(test_run_id=test_run_id)))
        test_run_ids.append(str(test_run_id))
    return test_run_ids



import logging
import re

from flask_login import current_user
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured

logger = logging.getLogger(__name__)


class TagsView(ModelViewConfigured):
    """View for :class:`Tags` table."""

    can_view_details = False

    column_list = ("value", "created_at", "created_by")
    form_excluded_columns = ("created_at",)
    column_searchable_list = ("value", "created_by")
    column_filters = ("value", "created_by")

    _tag_name_pattern = re.compile(r"^[0-9a-zA-Zа-яА-ЯёЁ_]+$")

    def on_model_change(self, form: Form, model: db.Tags, is_created: bool) -> None:
        if not self._tag_name_pattern.match(model.value):
            raise ValidationError("Unsupported symbols in tag name!")
        if not is_created:
            if current_user.login == model.created_by or current_user.role == db.Role.admin:
                return
            raise ValidationError("Only tag creator or administrator could edit it!")
        model.created_by = current_user.login

    def on_model_delete(self, model: db.Tags) -> None:
        if not (current_user.login == model.created_by or current_user.role == db.Role.admin):
            raise ValidationError("Only author or administrator could delete tags!")



from overhave.publication.settings import BaseGitPublisherSettings
from overhave.storage import FeatureTypeName
from overhave.transport import StashBranch, StashProject, StashRepository, StashReviewer, StashReviewerInfo


class NotSpecifiedFeatureTypeError(RuntimeError):
    """Exception for not specified reviewers relative to feature type."""


class OverhaveStashPublisherSettings(BaseGitPublisherSettings):
    """Settings for :class:`StashVersionPublisher`.

    This is a representation of BitBucket project parameters.
    Some pull-request parameters are also could be defined through these settings.
    """

    repository_name: str  # for example 'bdd-features'
    project_key: str  # for example 'PRJ'

    class Config:
        env_prefix = "OVERHAVE_STASH_"

    @property
    def repository(self) -> StashRepository:
        return StashRepository(slug=self.repository_name, project=StashProject(key=self.project_key))

    @property
    def target_branch(self) -> StashBranch:
        return StashBranch(id=self.default_target_branch_name, repository=self.repository)

    def get_reviewers(self, feature_type: FeatureTypeName) -> list[StashReviewer]:
        if self.feature_type_to_reviewers_mapping:
            reviewers = self.feature_type_to_reviewers_mapping.get(feature_type)
            if not reviewers:
                raise NotSpecifiedFeatureTypeError(
                    f"'{feature_type}' reviewers are not specified in 'feature_type_to_reviewers_mapping' dict!"
                )
        else:
            reviewers = self.default_reviewers
        return [StashReviewer(user=StashReviewerInfo(name=reviewer)) for reviewer in reviewers]



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field, TypeAdapter, model_validator


class BucketModel(BaseModel):
    """Model for boto3 client Bucket."""

    name: str = Field(alias="Name")
    created_at: datetime = Field(alias="CreationDate")


LIST_BUCKET_MODEL_ADAPTER: TypeAdapter[list[BucketModel]] = TypeAdapter(list[BucketModel])


class OwnerModel(BaseModel):
    """Model for boto3 client object owner."""

    name: str = Field(alias="DisplayName")
    owner_id: str = Field(alias="ID")


class ObjectModel(BaseModel):
    """Model for boto3 client object."""

    name: str = Field(alias="Key")
    modified_at: datetime = Field(alias="LastModified")
    etag: str = Field(alias="ETag")
    size: int = Field(alias="Size")
    storage_class: str = Field(alias="StorageClass")
    owner: OwnerModel = Field(alias="Owner")

    class Config:
        extra = "allow"


LIST_OBJECT_MODEL_ADAPTER: TypeAdapter[list[ObjectModel]] = TypeAdapter(list[ObjectModel])


class BaseObjectToDeletionModel(BaseModel):
    """Base model for boto3 client object deletion result."""

    name: str = Field(alias="Key")
    etag: str | None = Field(default=None, alias="VersionId")


class DeletedObjectModel(BaseObjectToDeletionModel):
    """Model for boto3 client deleted object."""

    marker: bool | None = Field(default=None, alias="DeleteMarker")
    marker_id: bool | None = Field(default=None, alias="DeleteMarkerVersionId")


class NotDeletedObjectModel(BaseObjectToDeletionModel):
    """Model for boto3 client not deleted object."""

    code: str = Field(alias="Code")
    message: str = Field(alias="Message")


class DeletionResultModel(BaseModel):
    """Model for boto3 client objects deletion result."""

    deleted: list[DeletedObjectModel] | None = Field(default=None, alias="Deleted")
    errors: list[NotDeletedObjectModel] | None = Field(default=None, alias="Errors")
    requester: str | None = Field(default=None, alias="RequestCharged")

    @model_validator(mode="before")
    def validate_results(cls, values: dict[str, Any]) -> dict[str, Any]:
        deleted = values.get("Deleted")
        errors = values.get("Errors")
        if deleted is None and errors is None:
            raise ValueError("At least one result field should be presented!")
        return values



import logging

import redis
import walrus

from overhave.metrics import BaseOverhaveMetricContainer
from overhave.transport.redis.objects import BaseRedisTask, RedisStream
from overhave.transport.redis.settings import BaseRedisSettings

logger = logging.getLogger(__name__)


class RedisProducer:
    """Class for producing tasks.

    Producer send tasks into Redis stream specified by ```mapping``.
    """

    def __init__(
        self,
        settings: BaseRedisSettings,
        mapping: dict[type[BaseRedisTask], RedisStream],
        database: walrus.Database,
        metric_container: BaseOverhaveMetricContainer,
    ):
        self._settings = settings
        self._streams = {task: database.Stream(stream.value) for task, stream in mapping.items()}
        self._mapping = mapping
        self._metric_container = metric_container

    def add_task(self, task: BaseRedisTask) -> bool:
        stream = self._streams[type(task)]
        logger.info("Added Redis task %s", task)
        try:
            stream.add(task.message)
            self._metric_container.produce_redis_task(task_type=self._mapping[type(task)].value)
            return True
        except redis.exceptions.ConnectionError:
            logger.exception("Could not add %s to Redis!", type(task).__name__)
            return False



import logging
import subprocess  # noqa: S404
from os import makedirs
from pathlib import Path
from uuid import uuid1

from overhave.db import TestReportStatus
from overhave.entities.archiver import ArchiveManager
from overhave.entities.report_manager.models import ReportPresenceResolution
from overhave.entities.settings import OverhaveFileSettings, OverhaveReportManagerSettings
from overhave.storage import ITestRunStorage
from overhave.transport import OverhaveS3Bucket, S3Manager

logger = logging.getLogger(__name__)


class ReportManager:
    """Class for Allure reports creation and management."""

    def __init__(
        self,
        settings: OverhaveReportManagerSettings,
        file_settings: OverhaveFileSettings,
        test_run_storage: ITestRunStorage,
        archive_manager: ArchiveManager,
        s3_manager: S3Manager,
    ) -> None:
        self._settings = settings
        self._file_settings = file_settings
        self._test_run_storage = test_run_storage
        self._archive_manager = archive_manager
        self._s3_manager = s3_manager

    def _generate_report(self, alluredir: Path, report_dir: Path) -> int | None:
        generation_cmd = (
            self._settings.allure_cmdline,
            "generate",
            f"{alluredir}/",
            "--output",
            report_dir.as_posix(),
            "--clean",
        )
        logger.debug("Allure report generation command: %s", " ".join(generation_cmd))
        makedirs(report_dir.as_posix(), exist_ok=True)
        try:
            return subprocess.run(  # noqa: S603
                generation_cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=self._settings.report_creation_timeout,
                check=True,
            ).returncode
        except (FileNotFoundError, subprocess.CalledProcessError):
            logger.exception("Error while generating Allure report!")
            return None

    def _process_generated_report(self, test_run_id: int, report_dir: Path) -> None:
        self._test_run_storage.set_report(run_id=test_run_id, status=TestReportStatus.GENERATED, report=report_dir.name)
        if not self._s3_manager.enabled:
            return
        zip_report = self._archive_manager.archive_path(path=report_dir, extension=self._settings.archive_extension)
        logger.info("Zip Allure report: %s", zip_report)
        upload_result = self._s3_manager.upload_file(file=zip_report, bucket=OverhaveS3Bucket.REPORTS)
        if not upload_result:
            return
        self._test_run_storage.set_report(run_id=test_run_id, status=TestReportStatus.SAVED)
        zip_report.unlink()

    def create_allure_report(self, test_run_id: int, results_dir: Path) -> None:
        report_dir = self._file_settings.tmp_reports_dir / uuid1().hex
        logger.debug("Allure report directory: %s", report_dir)

        report_generation_returncode = self._generate_report(alluredir=results_dir, report_dir=report_dir)
        if report_generation_returncode != 0:
            self._test_run_storage.set_report(run_id=test_run_id, status=TestReportStatus.GENERATION_FAILED)
            return
        logger.debug("Allure report successfully generated to directory: %s", report_dir.as_posix())
        self._process_generated_report(test_run_id=test_run_id, report_dir=report_dir)

    def get_report_precense_resolution(self, report: str, run_id: int) -> ReportPresenceResolution:  # noqa: C901
        report_index = Path(self._file_settings.tmp_reports_dir / report)
        report_dir = report_index.parent
        if report_dir.exists() and report_index.exists():
            return ReportPresenceResolution(exists=True)

        if not report_dir.exists():
            logger.warning("Report '%s' does not exist!", report_index.parent.name)
        if not report_index.exists():
            logger.warning("Report '%s' does not contain compiled files for HTML view!", report_index.parent.name)
        test_run = self._test_run_storage.get_testrun_model(run_id=run_id)
        if test_run is None:
            logger.warning("No one test run with id=%s exists!", run_id)
            return ReportPresenceResolution(exists=False)

        resolution = ReportPresenceResolution(
            exists=False, s3_enabled=self._s3_manager.enabled, report_status=test_run.report_status
        )
        if not test_run.report_status.has_report:
            logger.warning("TestRun with id=%s has not got report!", run_id)
            return resolution
        if not self._s3_manager.enabled:
            logger.info("S3Manager disabled, so could not download file from s3 cloud.")
            return resolution
        if test_run.report_status is TestReportStatus.GENERATED:
            logger.warning(
                "File %s after TestRun with id=%s has been generated, but still not uploaded!", test_run.report, run_id
            )
            return resolution
        zip_report_path = self._file_settings.tmp_reports_dir / (
            report_dir.name + f".{self._settings.archive_extension}"
        )
        zip_report_path.parent.mkdir(parents=True, exist_ok=True)
        download_success = self._s3_manager.download_file(
            filename=zip_report_path.name, dir_to_save=zip_report_path.parent, bucket=OverhaveS3Bucket.REPORTS.value
        )
        if not download_success:
            logger.error("Report archive '%s' is not available on s3 cloud!", zip_report_path.name)
            return resolution

        unpacked_report = self._archive_manager.unpack_path(
            path=zip_report_path, extension=self._settings.archive_extension
        )
        zip_report_path.unlink()
        logger.info("Unpacked Allure report: %s", unpacked_report)
        return ReportPresenceResolution(
            exists=True, s3_enabled=self._s3_manager.enabled, report_status=test_run.report_status
        )



import abc


class IFeatureValidator(abc.ABC):
    """Interface for features validation."""

    @abc.abstractmethod
    def validate(self, raise_if_nullable_id: bool = False, pull_repository: bool = False) -> None:
        pass



import json
import logging
import re
from functools import cached_property
from pathlib import Path
from typing import Any, Callable, cast

import flask
import werkzeug
from flask_admin import expose
from flask_admin.model import InlineFormAdmin
from flask_login import current_user
from markupsafe import Markup
from pytest_bdd.types import STEP_TYPES
from wtforms import Field, Form, TextAreaField, ValidationError
from wtforms.widgets import HiddenInput

from overhave import db
from overhave.admin.views.base import ModelViewConfigured
from overhave.factory import IAdminFactory, get_admin_factory, get_test_execution_factory
from overhave.pytest_plugin import get_proxy_manager
from overhave.storage import FeatureTypeName
from overhave.test_execution import BddStepModel, StepTypeName
from overhave.transport import TestRunData, TestRunTask
from overhave.utils import get_current_time

logger = logging.getLogger(__name__)

_SCENARIO_PREFIX = "scenario-0"


class ScenarioTextWidget(HiddenInput):
    """Widget to override scenario view."""

    def __call__(self, field: Field, **kwargs: Any) -> Any:
        widget_name = type(self).__name__
        logger.debug("%s field id: '%s'", widget_name, field.id)
        inputs = super().__call__(field, **kwargs)
        logger.debug("%s inputs: '%s'", widget_name, inputs)
        return inputs + Markup(f'<div class="editor" ref="{field.id}" id="{field.id}-editor"></div>')


class ScenarioTextAreaField(TextAreaField):
    """Field to override scenario view."""

    widget = ScenarioTextWidget()


class ScenarioInlineModelForm(InlineFormAdmin):
    """Form to override scenario view."""

    form_overrides = dict(text=ScenarioTextAreaField)
    form_excluded_columns = ("created_at", "test_runs")


class FactoryViewUtilsMixin:
    """Mixin for :class:`FeatureView`. Extra methods for working with cached instance of :class:`IAdminFactory`."""

    _task_pattern = re.compile(r"\w+-\d+")
    _file_path_pattern = re.compile(r"^[0-9a-zA-Zа-яА-ЯёЁ_/\\ ]{8,}")

    @classmethod
    def _validate_tasks(cls, tasks: list[str]) -> None:
        for task in tasks:
            if cls._task_pattern.match(task):
                continue
            raise ValidationError(
                f"Incorrect format of task specification: '{task}'! "
                "Supported: <PROJECT>-<NUMBER>, for example 'PRJ-1234'."
            )

    def _make_file_path(self, file_path: str) -> str:
        if file_path.rstrip().endswith(self.feature_suffix):
            file_path = file_path.removesuffix(self.feature_suffix)
        if not self._file_path_pattern.match(file_path):
            raise ValidationError(
                f"Incorrect format of file path specification: '{file_path}'! "
                f"Supported pattern: {self._file_path_pattern.pattern}, for example 'my_folder/my_filename(.feature)'."
                " At least 8 characters long."
            )
        prepared_file_path = file_path.replace(" ", "").lstrip("/")
        path = Path(prepared_file_path).with_suffix(self.feature_suffix).as_posix()
        logger.debug("Processed feature file path: '%s'", path)
        return path

    @property
    def task_tracker_url(self) -> str | None:
        task_tracker_url_value = get_admin_factory().context.project_settings.task_tracker_url
        if task_tracker_url_value is not None:
            return str(task_tracker_url_value)
        return None

    @cached_property
    def feature_suffix(self) -> str:
        return get_admin_factory().context.file_settings.feature_suffix

    @staticmethod
    def _get_feature_type_steps(factory: IAdminFactory, feature_type: FeatureTypeName) -> list[BddStepModel]:
        return factory.step_collector.get_steps(feature_type) or []

    @cached_property
    def get_bdd_steps(self) -> dict[FeatureTypeName, dict[StepTypeName, list[BddStepModel]]]:
        factory = get_admin_factory()
        return {
            feature_type: {
                StepTypeName(step_type): [
                    step
                    for step in self._get_feature_type_steps(factory=factory, feature_type=feature_type)
                    if step.type == step_type
                ]
                for step_type in STEP_TYPES
            }
            for feature_type in factory.feature_extractor.feature_types
        }


class FeatureView(ModelViewConfigured, FactoryViewUtilsMixin):
    """View for :class:`Feature` table."""

    can_view_details = False

    inline_models = (ScenarioInlineModelForm(db.Scenario),)

    create_template = "feature_create.html"
    edit_template = "feature_edit.html"

    column_list = (
        "id",
        "name",
        "feature_type",
        "file_path",
        "severity",
        "feature_tags",
        "task",
        "last_edited_at",
        "last_edited_by",
        "released",
    )
    form_excluded_columns = (
        "created_at",
        "last_edited_by",
        "last_edited_at",
        "released",
        "versions",
        "feature_tags.value",
    )

    column_searchable_list = (
        "id",
        "name",
        "task",
        "author",
        "file_path",
        "last_edited_by",
        "feature_tags.value",
    )
    column_filters = (
        "id",
        "name",
        "feature_type",
        "last_edited_by",
        "author",
        "created_at",
        "last_edited_at",
        "feature_tags.value",
        "severity",
    )
    column_sortable_list = ("id", "name", "author", "last_edited_by", "severity")
    column_labels = {
        "file_path": "File",
        "feature_tags": "Tags",
        "feature_type": "Type",
        "name": "Name",
        "task": "Tasks",
    }
    column_descriptions = {
        "feature_type": "Feature type / root directory",
        "name": "Feature header for business scenarios",
        "file_path": "Path for saving current file",
        "severity": "Feature severity level",
        "feature_tags": "Special tags for custom marking",
        "task": "Tracker tasks converted to web links",
        "last_edited_by": "Last editor of scenarios set",
        "author": "Author of scenarios set",
        "released": "Last version publishing status",
    }
    form_columns = (
        "feature_type",
        "name",
        "file_path",
        "severity",
        "feature_tags",
        "task",
        "scenario",
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def on_model_change(self, form, model: db.Feature, is_created) -> None:  # type: ignore
        self._validate_tasks(model.task)
        model.file_path = self._make_file_path(model.file_path)
        if is_created:
            model.author = current_user.login
        model.last_edited_by = current_user.login
        model.last_edited_at = get_current_time()
        model.released = False

    def on_model_delete(self, model) -> None:  # type: ignore
        if not (current_user.login == model.author or current_user.role == db.Role.admin):
            raise ValidationError("Only feature author or administrator could delete feature!")

    @staticmethod
    def _run_test(data: dict[str, Any], rendered: werkzeug.Response) -> werkzeug.Response:
        scenario_id = data.get(f"{_SCENARIO_PREFIX}-id")
        scenario_text = data.get(f"{_SCENARIO_PREFIX}-text")
        if not scenario_id or not scenario_text:
            flask.flash("Scenario information not requested.", category="error")
            return rendered
        factory = get_admin_factory()
        with db.create_session() as session:
            scenario = factory.scenario_storage.scenario_model_by_id(session=session, scenario_id=int(scenario_id))
        test_run_id = factory.test_run_storage.create_testrun(scenario_id=scenario.id, executed_by=current_user.login)
        if not factory.context.admin_settings.consumer_based:
            proxy_manager = get_proxy_manager()
            test_execution_factory = get_test_execution_factory()
            proxy_manager.clear_factory()
            proxy_manager.set_factory(test_execution_factory)
            factory.threadpool.apply_async(get_test_execution_factory().test_executor.execute_test, args=(test_run_id,))
        if factory.context.admin_settings.consumer_based and not factory.redis_producer.add_task(
            TestRunTask(data=TestRunData(test_run_id=test_run_id))
        ):
            flask.flash("Problems with Redis service! TestRunTask has not been sent.", category="error")
            return rendered
        logger.debug("Redirect to TestRun details view with test_run_id='%s'...", test_run_id)
        return flask.redirect(flask.url_for("testrun.details_view", id=test_run_id))

    @expose("/edit/", methods=("GET", "POST"))
    def edit_view(self) -> werkzeug.Response:
        rendered: werkzeug.Response = super().edit_view()
        if flask.request.method != "POST":
            return rendered

        data = flask.request.form
        logger.debug("Request data:\n%s", json.dumps(data))
        run_scenario_action = data.get("run")
        if not run_scenario_action:
            logger.debug("Show rendered EditView")
            return rendered

        logger.debug("Process feature 'RUN' request")
        if data["task"]:
            tasks = data["task"].split(",")
            self._validate_tasks(tasks=tasks)

        mutable_data = dict(data)
        mutable_data["file_path"] = self._make_file_path(data["file_path"])

        return self._run_test(data, rendered)

    @staticmethod
    def _form_remove_validators(form_change: Callable[..., Form]) -> Callable[..., Form]:
        admin_settings = get_admin_factory().context.admin_settings
        if not admin_settings.strict_feature_tasks:
            form_change.task.kwargs["validators"] = []  # type: ignore[attr-defined]
        return form_change

    def get_create_form(self) -> Callable[..., Form]:
        form_create = cast(Callable[..., Form], super().get_create_form())
        return self._form_remove_validators(form_change=form_create)

    def get_edit_form(self) -> Callable[..., Form]:
        form_edit = cast(Callable[..., Form], super().get_edit_form())
        return self._form_remove_validators(form_change=form_edit)



import logging

from gitlab.exceptions import GitlabAuthenticationError
from gitlab.v4.objects import ProjectMergeRequest

from overhave.transport.http import BaseHttpClient
from overhave.transport.http.base_client import BaseHttpClientException, HttpClientValidationError
from overhave.transport.http.gitlab_client.models import GitlabMrCreationResponse, GitlabMrRequest
from overhave.transport.http.gitlab_client.settings import OverhaveGitlabClientSettings
from overhave.transport.http.gitlab_client.utils import get_gitlab_python_client

logger = logging.getLogger(__name__)


class BaseGitlabHttpClientException(BaseHttpClientException):
    """Base exception for :class:`GitlabHttpClient`."""


class GitlabInvalidTokenError(BaseGitlabHttpClientException):
    """Exception for situation with invalid token or gitlab url."""


class GitlabHttpClient(BaseHttpClient[OverhaveGitlabClientSettings]):
    """Client for communication with remote Gitlab server."""

    def send_merge_request(
        self, repository_id: str, merge_request: GitlabMrRequest, token: str
    ) -> GitlabMrCreationResponse:
        gitlab_python_client = get_gitlab_python_client(
            url=self._settings.url,
            token_type=self._settings.token_type,
            token=token,
        )
        project = gitlab_python_client.projects.get(repository_id, lazy=True)
        try:
            result = project.mergerequests.create(merge_request.model_dump(by_alias=True))
            logger.debug("Response from GitLab: %s", result)
            if isinstance(result, ProjectMergeRequest):
                return GitlabMrCreationResponse(created_at=result.created_at, web_url=result.web_url)
            raise HttpClientValidationError("Unsupported response: %s", result)
        except GitlabAuthenticationError as e:
            logging.exception("Please verify your token or URL! Maybe they are invalid")
            raise GitlabInvalidTokenError("Please verify your token or URL! Maybe they are invalid") from e



import re
from functools import cached_property
from typing import Mapping, Pattern

import allure_commons.types
import pytest
from _pytest.mark.structures import MarkDecorator
from pydantic import BaseModel


class TagEvaluationResult(BaseModel):
    """Class for tag evaluation result."""

    marker: MarkDecorator
    url: str | None
    link_type: str

    class Config:
        arbitrary_types_allowed = True


class BaseOverhaveTagControllerException(Exception):
    """Base exception for :class:`OverhaveTagController`."""


class NoReasonForMarkDecoratorError(Exception):
    """Exception for situation with nullable reason in tag."""


class OverhaveTagController:
    """Class for mark decorators setup via special Overhave tags."""

    @staticmethod
    def _get_tag_pattern(keyword: str) -> Pattern[str]:
        return re.compile(rf"\b({keyword})+(\(+[^@()]+\)+)?\b")

    @cached_property
    def _tag_to_evaluation_result_mapping(
        self,
    ) -> Mapping[Pattern[str], tuple[MarkDecorator, allure_commons.types.LinkType]]:
        return {
            self._get_tag_pattern("disabled"): (
                pytest.mark.skip,
                allure_commons.types.LinkType.LINK,
            ),
            self._get_tag_pattern("xfail"): (
                pytest.mark.xfail,
                allure_commons.types.LinkType.ISSUE,
            ),
        }

    @cached_property
    def _reason_pattern(self) -> Pattern[str]:
        return re.compile(r"\w+\((?P<reason>[^@()]+)\)")

    @cached_property
    def _url_pattern(self) -> Pattern[str]:
        return re.compile(r"(?P<url>\b\w+://[^\s\"\']+\b)")

    def _get_url(self, reason: str) -> str | None:
        searched = self._url_pattern.search(reason)
        if searched is not None:
            return searched.group("url")
        return None

    def get_suitable_pattern(self, name: str) -> Pattern[str] | None:
        for pattern in self._tag_to_evaluation_result_mapping:
            result = pattern.match(name)
            if result is not None:
                return pattern
        return None

    def evaluate_tag(self, name: str, pattern: Pattern[str]) -> TagEvaluationResult:
        reason_error = NoReasonForMarkDecoratorError(
            f"Tag '{name}' has been used without reason! Please, setup reason using round brackets `(`, `)`\n"
            "For example: `@disabled(TODO: https://tracker.mydomain.com/browse/PRJ-333)`"
        )
        match = self._reason_pattern.match(name)
        if match is None:
            raise reason_error
        reason = match.group("reason")
        if reason is None:
            raise reason_error
        kwargs = {"reason": reason}
        mark_decorator, link_type = self._tag_to_evaluation_result_mapping[pattern]
        return TagEvaluationResult(marker=mark_decorator(**kwargs), url=self._get_url(reason), link_type=link_type)



import logging

from prometheus_client import CollectorRegistry, Counter

logger = logging.getLogger(__name__)


class BaseOverhaveMetricContainer:
    """Base prometheus overhave metric container. Contains common metrics."""

    def __init__(self, registry: CollectorRegistry):
        self.registry = registry
        self._init_redis_metrics()

    def _init_redis_metrics(self) -> None:
        self.produced_redis_tasks = Counter(
            "produced_redis_tasks",
            "How many redis tasks have been produced",
            labelnames=("task_type",),
            registry=self.registry,
        )
        self.consumed_redis_tasks = Counter(
            "consumed_redis_tasks",
            "How many redis tasks have been consumed",
            labelnames=("task_type",),
            registry=self.registry,
        )

    def produce_redis_task(self, task_type: str) -> None:
        self.produced_redis_tasks.labels(task_type=task_type).inc()

    def consume_redis_task(self, task_type: str) -> None:
        self.consumed_redis_tasks.labels(task_type=task_type).inc()


class TestRunOverhaveMetricContainer(BaseOverhaveMetricContainer):
    """Overhave prometheus metric container for test runs."""

    __test__ = False

    def __init__(self, registry: CollectorRegistry):
        super().__init__(registry=registry)
        self._init_test_run_metrics()

    def _init_test_run_metrics(self) -> None:
        self.test_run_tasks_statuses = Counter(
            "test_run_tasks_statuses",
            "Counter for test run statuses",
            labelnames=("status",),
            registry=self.registry,
        )

    def add_test_run_status(self, status: str) -> None:
        self.test_run_tasks_statuses.labels(status=status).inc()


class EmulationRunOverhaveMetricContainer(BaseOverhaveMetricContainer):
    """Overhave prometheus metric container for emulation runs."""

    def __init__(self, registry: CollectorRegistry):
        super().__init__(registry=registry)
        self._init_emulation_metrics()

    def _init_emulation_metrics(self) -> None:
        self.emulation_tasks_statuses = Counter(
            "emulation_tasks_statuses",
            "Counter for emulation statuses",
            labelnames=("status", "port"),
            registry=self.registry,
        )

    def add_emulation_task_status(self, status: str, port: int | None = None) -> None:
        self.emulation_tasks_statuses.labels(status=status, port=port).inc()


class PublicationOverhaveMetricContainer(BaseOverhaveMetricContainer):
    """Overhave prometheus metric container for publications."""

    def __init__(self, registry: CollectorRegistry):
        super().__init__(registry=registry)
        self._init_publication_metrics()

    def _init_publication_metrics(self) -> None:
        self.publication_tasks_statuses = Counter(
            "publication_tasks_statuses",
            "Counter for publication run statuses",
            labelnames=("status",),
            registry=self.registry,
        )

    def add_publication_task_status(self, status: str) -> None:
        self.publication_tasks_statuses.labels(status=status).inc()



import logging
from dataclasses import dataclass

from flask import redirect
from flask_login import LoginManager
from werkzeug import Response

from overhave import db
from overhave.storage import ISystemUserStorage, SystemUserModel

logger = logging.getLogger(__name__)


class UnauthorizedUserError(Exception):
    """Raises when user is not authorized and trying to get user fields such as role or login."""


@dataclass(frozen=True)
class AdminPanelUser:
    """SystemUserModel wrapper for flask_login."""

    user_data: SystemUserModel | None

    @property
    def is_authenticated(self) -> bool:
        return self.user_data is not None

    @property
    def is_active(self) -> bool:
        return True

    @property
    def is_anonymous(self) -> bool:
        return self.user_data is None

    @property
    def _authorized_user(self) -> SystemUserModel:
        if self.user_data is None:
            raise UnauthorizedUserError("User is not authorized!")
        return self.user_data

    def get_id(self) -> int:
        return self._authorized_user.id

    @property
    def login(self) -> str:
        return self._authorized_user.login

    @property
    def role(self) -> db.Role:
        return self._authorized_user.role

    def __unicode__(self) -> str:
        if self.user_data is not None:
            return self.user_data.login
        return "anonymous"


class FlaskLoginManager(LoginManager):
    """Custom Flask LoginManager for user acknowledgement."""

    def __init__(self, system_user_storage: ISystemUserStorage, login_view: str) -> None:
        super().__init__()
        self.login_view = login_view
        self.user_loader(self._get_user)
        self.unauthorized_handler(self._unathorized_response)

        self._system_user_storage = system_user_storage

    def _get_user(self, user_id: int) -> AdminPanelUser:
        logger.info("Get user by id=%s...", user_id)
        return AdminPanelUser(user_data=self._system_user_storage.get_user_model(user_id=user_id))

    def _unathorized_response(self) -> Response:
        return redirect(f"/{self.login_view}")



from datetime import datetime
from typing import NewType

import allure
from pydantic.main import BaseModel
from pydantic.types import SecretStr

from overhave import db
from overhave.db import DraftStatus, Role, TestReportStatus, TestRunStatus

FeatureTypeName = NewType("FeatureTypeName", str)
TestUserSpecification = NewType("TestUserSpecification", dict[str, str | None])


class _SqlAlchemyOrmModel(BaseModel):
    """:class:`BaseModel` with enabled `orm_mode`."""

    class Config:
        from_attributes = True


class SystemUserModel(_SqlAlchemyOrmModel):
    """Model for :class:`UserRole`."""

    id: int
    login: str
    password: SecretStr | None
    role: Role

    def get_id(self) -> int:
        return self.id

    def __unicode__(self) -> str:
        return self.login


class FeatureTypeModel(_SqlAlchemyOrmModel):
    """Model for :class:`FeatureType` row."""

    id: int
    name: FeatureTypeName


class TagModel(_SqlAlchemyOrmModel):
    """Model for :class:`Tags` row."""

    id: int
    value: str
    created_by: str


class FeatureModel(_SqlAlchemyOrmModel):
    """Model for :class:`Feature` row."""

    id: int
    created_at: datetime
    name: str
    author: str
    type_id: int
    last_edited_by: str
    last_edited_at: datetime
    task: list[str]
    file_path: str
    released: bool
    severity: allure.severity_level

    feature_type: FeatureTypeModel
    feature_tags: list[TagModel]


class ScenarioModel(_SqlAlchemyOrmModel):
    """Model for :class:`Scenario` row."""

    id: int
    text: str
    feature_id: int


class TestRunModel(_SqlAlchemyOrmModel):
    """Model for :class:`TestRun` row."""

    __test__ = False

    id: int
    created_at: datetime
    name: str
    executed_by: str
    start: datetime | None
    end: datetime | None
    status: TestRunStatus
    report_status: TestReportStatus
    report: str | None
    traceback: str | None
    scenario_id: int


class DraftModel(_SqlAlchemyOrmModel):
    """Model for :class:`Draft` row."""

    id: int
    feature_id: int
    test_run_id: int
    pr_url: str | None
    published_by: str
    published_at: datetime | None
    traceback: str | None
    status: DraftStatus


class TestExecutorContext(BaseModel):
    """Context model for test execution."""

    __test__ = False

    feature: FeatureModel
    scenario: ScenarioModel
    test_run: TestRunModel


class PublisherContext(TestExecutorContext):
    """Context model for version publishing."""

    draft: DraftModel
    target_branch: str


class TestUserModel(_SqlAlchemyOrmModel):
    """Model for :class:`TestUser` row."""

    __test__ = False

    id: int
    created_at: datetime
    key: str
    name: str
    created_by: str
    specification: TestUserSpecification
    feature_type_id: int
    feature_type: FeatureTypeModel
    allow_update: bool
    changed_at: datetime


class EmulationModel(_SqlAlchemyOrmModel):
    """Model for :class:`Emulation` row."""

    id: int
    command: str
    test_user: TestUserModel


class EmulationRunModel(_SqlAlchemyOrmModel):
    """Model for :class:`EmulationRun` row."""

    id: int
    emulation_id: int
    changed_at: datetime
    status: db.EmulationStatus
    port: int | None
    initiated_by: str
    emulation: EmulationModel



import logging
from typing import cast

from pydantic.main import BaseModel

from overhave.publication.gitlab.tokenizer.settings import TokenizerClientSettings
from overhave.transport.http import BaseHttpClient
from overhave.transport.http.base_client import HttpMethod

logger = logging.getLogger(__name__)


class TokenizerResponse(BaseModel):
    """Response from service tokenizer with token."""

    token: str


class TokenizerRequestParamsModel(BaseModel):
    """Request for service tokenizer."""

    initiator: str
    id: int
    remote_key: str

    def get_request_params(self, remote_key_name: str) -> dict[str, int | str]:
        return {"initiator": self.initiator, "id": self.id, remote_key_name: self.remote_key}


class BaseTokenizerException(Exception):
    """Base tokenizer exception."""


class InvalidUrlException(BaseTokenizerException):
    """Exception for case when tokenizer url is None."""


class InvalidRemoteKeyNameException(BaseTokenizerException):
    """Exception for case when tokenizer remote key name is None."""


class TokenizerClient(BaseHttpClient[TokenizerClientSettings]):
    """Client for sending requests for getting tokens for gitlab."""

    def __init__(self, settings: TokenizerClientSettings):
        super().__init__(settings)
        self._settings = settings

    def get_token(self, draft_id: int) -> TokenizerResponse:
        params_model = TokenizerRequestParamsModel(
            initiator=self._settings.initiator, id=draft_id, remote_key=self._settings.remote_key
        )
        if self._settings.url is None:
            raise InvalidUrlException("Please check tokenizer url! Value is invalid!")
        if self._settings.remote_key_name is None:
            raise InvalidRemoteKeyNameException("Please check remote key name! Value is invalid!")
        response = self._make_request(
            HttpMethod.POST,
            self._settings.url,
            params=params_model.get_request_params(self._settings.remote_key_name),
        )
        logger.debug("Response with token: %s", response)
        return cast(TokenizerResponse, self._parse_or_raise(response, TokenizerResponse))



import abc

import sqlalchemy.orm as so

from overhave import db
from overhave.storage.converters import TagModel


class IFeatureTagStorage(abc.ABC):
    """Abstract class for feature tags storage."""

    @staticmethod
    @abc.abstractmethod
    def get_tag_by_value(value: str) -> TagModel | None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_tags_like_value(value: str) -> list[TagModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_or_create_tag(session: so.Session, value: str, created_by: str) -> db.Tags:
        pass


class FeatureTagStorage(IFeatureTagStorage):
    """Class for feature tags storage."""

    @staticmethod
    def get_tag_by_value(value: str) -> TagModel | None:
        with db.create_session() as session:
            tag = session.query(db.Tags).filter(db.Tags.value == value).one_or_none()
            if tag is not None:
                return TagModel.model_validate(tag)
            return None

    @staticmethod
    def get_tags_like_value(value: str) -> list[TagModel]:
        with db.create_session() as session:
            db_tags = session.query(db.Tags).filter(db.Tags.value.like(value)).all()
            return [TagModel.model_validate(tag) for tag in db_tags]

    @staticmethod
    def get_or_create_tag(session: so.Session, value: str, created_by: str) -> db.Tags:
        tag = session.query(db.Tags).filter(db.Tags.value == value).one_or_none()
        if tag is None:
            tag = db.Tags(value=value, created_by=created_by)
            session.add(tag)
        return tag



import httpx
from pydantic import field_validator
from pydantic_core.core_schema import ValidationInfo

from overhave.transport.http import BaseHttpClientSettings
from overhave.utils import make_url


class TokenizerClientSettings(BaseHttpClientSettings):
    """Important environments and settings for :class:`TokenizerClient`."""

    enabled: bool = False
    url: httpx.URL | None = None  # type: ignore
    initiator: str = "Overhave"
    remote_key: str | None = None
    remote_key_name: str | None = None

    class Config:
        env_prefix = "OVERHAVE_GITLAB_TOKENIZER_"

    @field_validator("url", mode="before")
    def make_url(cls, v: str | None) -> httpx.URL | None:
        return make_url(v)

    @field_validator("url", "remote_key", "remote_key_name", mode="after")
    def validate_remote_key_and_initiator(cls, v: str | None, values: ValidationInfo) -> str | None:
        if values.data.get("enabled") and not isinstance(v, str):
            raise ValueError("Please verify that url, remote_key and remote_key_name variables are not nullable!")
        return v



import sqlalchemy as sa
import sqlalchemy_utils as sau
import typer
from alembic.config import Config
from sqlalchemy.exc import OperationalError

from overhave import db
from overhave.base_settings import DataBaseSettings


def create_schema(config: Config) -> None:
    typer.echo("Creating...")
    config.attributes["metadata"].create_all(bind=db.metadata.engine)
    typer.secho("Completed.", fg="green")


def drop_schema(config: Config) -> None:
    typer.echo("Dropping...")
    meta = config.attributes["metadata"]
    engine: sa.Engine = config.attributes["engine"]

    connection: sa.Connection = engine.connect()
    for table in meta.tables:
        connection.execute(sa.text(f'DROP TABLE IF EXISTS "{table}" CASCADE'))
    connection.execute(sa.text("DROP TABLE IF EXISTS alembic_version"))

    meta.drop_all(engine)
    typer.secho("Completed.", fg="green")


def _ensure_database_exists(db_url: sa.URL) -> None:
    try:
        if not sau.database_exists(db_url):
            sau.create_database(db_url)
    except OperationalError as e:
        typer.echo(e)
        typer.echo("Catched error when trying to check database existence!")


def set_config_to_context(context: typer.Context, settings: DataBaseSettings) -> None:
    """Set Alembic config to Typer context for easy operations and migrations ability."""
    settings.setup_engine()
    _ensure_database_exists(db.metadata.engine.url)

    config = Config()
    config.attributes["engine"] = db.metadata.engine
    config.attributes["metadata"] = db.metadata
    context.obj = config



import logging

import httpx

from overhave import db
from overhave.entities import GitRepositoryInitializer
from overhave.metrics import PublicationOverhaveMetricContainer
from overhave.publication.git_publisher import GitVersionPublisher
from overhave.publication.stash.settings import OverhaveStashPublisherSettings
from overhave.scenario import FileManager, OverhaveProjectSettings
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.transport import (
    StashBranch,
    StashErrorResponse,
    StashHttpClient,
    StashHttpClientConflictError,
    StashPrCreationResponse,
    StashPrRequest,
)

logger = logging.getLogger(__name__)


class StashVersionPublisher(GitVersionPublisher[OverhaveStashPublisherSettings]):
    """Class for feature version's pull requests management relative to Atlassian Stash API."""

    def __init__(
        self,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
        git_initializer: GitRepositoryInitializer,
        stash_publisher_settings: OverhaveStashPublisherSettings,
        stash_client: StashHttpClient,
        metric_container: PublicationOverhaveMetricContainer,
    ):
        super().__init__(
            project_settings=project_settings,
            feature_storage=feature_storage,
            scenario_storage=scenario_storage,
            test_run_storage=test_run_storage,
            draft_storage=draft_storage,
            file_manager=file_manager,
            git_initializer=git_initializer,
            git_publisher_settings=stash_publisher_settings,
            metric_container=metric_container,
        )
        self._client = stash_client

    def publish_version(self, draft_id: int) -> db.DraftStatus:
        context = self._prepare_publisher_context(draft_id)
        if context is None:
            self._draft_storage.set_draft_status(draft_id=draft_id, status=db.DraftStatus.INTERNAL_ERROR)
            return db.DraftStatus.INTERNAL_ERROR
        pull_request = StashPrRequest(
            title=context.feature.name,
            description=self._compile_publication_description(context),
            open=True,
            fromRef=StashBranch(id=context.target_branch, repository=self._git_publisher_settings.repository),
            toRef=self._git_publisher_settings.target_branch,
            reviewers=self._git_publisher_settings.get_reviewers(feature_type=context.feature.feature_type.name),
        )
        logger.info("Prepared pull-request: %s", pull_request.model_dump_json(by_alias=True))
        try:
            response = self._client.send_pull_request(pull_request)
            if isinstance(response, StashPrCreationResponse):
                self._draft_storage.save_response_as_created(
                    draft_id=draft_id,
                    pr_url=response.get_pr_url(),
                    published_at=response.created_date,
                )
                return db.DraftStatus.CREATED
            if isinstance(response, StashErrorResponse) and response.duplicate:
                self._draft_storage.save_response_as_duplicate(
                    draft_id=context.draft.id, feature_id=context.feature.id, traceback=None
                )
                return db.DraftStatus.DUPLICATE
            self._draft_storage.set_draft_status(draft_id, db.DraftStatus.INTERNAL_ERROR)
            logger.error("Gotten error response from Stash: %s", response)
            return db.DraftStatus.INTERNAL_ERROR

        except StashHttpClientConflictError as err:
            logger.exception("Gotten conflict. Try to return last pull-request for Draft with id=%s...", draft_id)
            self._draft_storage.save_response_as_duplicate(
                draft_id=context.draft.id, feature_id=context.feature.id, traceback=str(err)
            )
            return db.DraftStatus.DUPLICATE
        except httpx.HTTPError as err:
            self._draft_storage.set_draft_status(draft_id, db.DraftStatus.INTERNAL_ERROR, str(err))
            logger.exception("Got HTTP error while trying to sent pull-request!")
            return db.DraftStatus.INTERNAL_ERROR



import logging
from pathlib import Path

from overhave.entities import BaseFileExtractor, GitRepositoryInitializer, OverhaveFileSettings
from overhave.scenario.parser import NullableFeatureIdError, ScenarioParser, StrictFeatureParsingError
from overhave.scenario.validator.abstract import IFeatureValidator
from overhave.scenario.validator.duplicate_id_mixin import FeatureDuplicatedIdValidationMixin
from overhave.scenario.validator.errors import FeaturesWithoutIDPresenceError, IncorrectFeaturesPresenceError

logger = logging.getLogger(__name__)


class FeatureValidator(IFeatureValidator, BaseFileExtractor, FeatureDuplicatedIdValidationMixin):
    """Class for features validation."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        scenario_parser: ScenarioParser,
        git_initializer: GitRepositoryInitializer,
    ):
        super().__init__(extenstion=file_settings.feature_suffix)
        self._file_settings = file_settings
        self._scenario_parser = scenario_parser
        self._git_initializer = git_initializer

        self._scenario_parser.set_strict_mode(True)

        self._nullable_id_features: list[Path] = []
        self._incorrect_features: list[Path] = []

    def validate(self, raise_if_nullable_id: bool = False, pull_repository: bool = False) -> None:
        if pull_repository:
            self._git_initializer.pull()
        logger.info("Start validation...")
        feature_paths = self._extract_recursively(self._file_settings.features_dir)
        for feature_path in feature_paths:
            logger.info("Read feature from file %s...", feature_path.as_posix())
            try:
                feature_info = self._scenario_parser.parse(feature_path.read_text())
                self._save_to_feature_id_to_path_mapping(feature_path=feature_path, feature_id=feature_info.id)
            except NullableFeatureIdError:
                logger.warning("Feature has not got suitable Overhave ID!")
                self._nullable_id_features.append(feature_path)
            except StrictFeatureParsingError:
                logger.exception("Feature has incorrect format!")
                self._incorrect_features.append(feature_path)
            else:
                logger.info("Feature successfully parsed: %s", feature_info)
        if self._incorrect_features:
            raise IncorrectFeaturesPresenceError(
                f"Features with incorrect format: {[x.as_posix() for x in self._incorrect_features]}"
            )
        if raise_if_nullable_id and self._nullable_id_features:
            raise FeaturesWithoutIDPresenceError(
                f"Features without IDs: {[x.as_posix() for x in self._nullable_id_features]}"
            )
        self._validate_duplicate_ids()
        logger.info("All features are correct")



from dataclasses import dataclass

from overhave.db import TestReportStatus


@dataclass(frozen=True)
class ReportPresenceResolution:
    """Model for report presence resolution result."""

    exists: bool

    report_status: TestReportStatus | None = None
    s3_enabled: bool = False

    @property
    def not_ready(self) -> bool:
        return self.s3_enabled and not self.exists and self.report_status is TestReportStatus.GENERATED



from markupsafe import Markup

from overhave.db import TestRunStatus


def get_button_class_by_status(status: str) -> str:
    enum_member = TestRunStatus[status]
    if enum_member is TestRunStatus.SUCCESS:
        return "success-btn"
    return "default-btn"


def get_testrun_details_link(test_run_id: int | str) -> str:
    return f"/testrun/details/?id={test_run_id}"


def get_report_index_link(report: str) -> str:
    return f"/reports/{report}/index.html"


def get_feature_link_markup(feature_id: int | str, feature_name: str) -> Markup:
    return Markup(f"<a href='/feature/edit/?id={feature_id}'>{feature_name}</a>")



import abc

import sqlalchemy.orm as so

from overhave import db


class ISystemUserGroupStorage(abc.ABC):
    """Abstract class for system user storage."""

    @staticmethod
    @abc.abstractmethod
    def has_any_group(session: so.Session, user_groups: list[str]) -> bool:
        pass


class SystemUserGroupStorage(ISystemUserGroupStorage):
    """Class for system user storage."""

    @staticmethod
    def has_any_group(session: so.Session, user_groups: list[str]) -> bool:
        return session.query(db.GroupRole).filter(db.GroupRole.group.in_(user_groups)).first() is not None



import abc

import sqlalchemy as sa
import sqlalchemy.orm as so

from overhave import db
from overhave.storage import TestUserModel, TestUserSpecification
from overhave.utils import get_current_time


class BaseTestUserStorageException(Exception):
    """Base exception for :class:`FeatureStorage`."""


class TestUserDoesNotExistError(BaseTestUserStorageException):
    """Error for situation when test user not found."""

    __test__ = False


class TestUserUpdatingNotAllowedError(BaseTestUserStorageException):
    """Error for situation when test user has allow_update=False."""

    __test__ = False


class ITestUserStorage(abc.ABC):
    """Abstract class for Test User storage."""

    @staticmethod
    @abc.abstractmethod
    def get_testuser_model_by_id(user_id: int) -> TestUserModel | None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_testuser_model_by_key(key: str) -> TestUserModel | None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_test_users_by_feature_type_name(
        session: so.Session, feature_type_id: int, allow_update: bool
    ) -> list[TestUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_test_user_specification(user_id: int, specification: TestUserSpecification) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def delete_test_user(user_id: int) -> None:
        pass


class TestUserStorage(ITestUserStorage):
    """Class for Test User storage."""

    @staticmethod
    def get_testuser_model_by_id(user_id: int) -> TestUserModel | None:
        with db.create_session() as session:
            user = session.get(db.TestUser, user_id)
            if user is not None:
                return TestUserModel.model_validate(user)
            return None

    @staticmethod
    def get_testuser_model_by_key(key: str) -> TestUserModel | None:
        with db.create_session() as session:
            user: db.TestUser | None = session.query(db.TestUser).filter(db.TestUser.key == key).one_or_none()
            if user is not None:
                return TestUserModel.model_validate(user)
            return None

    @staticmethod
    def get_test_users_by_feature_type_name(
        session: so.Session, feature_type_id: int, allow_update: bool
    ) -> list[TestUserModel]:
        db_users = (
            session.query(db.TestUser)
            .filter(db.TestUser.feature_type_id == feature_type_id, db.TestUser.allow_update.is_(allow_update))
            .all()
        )
        return [TestUserModel.model_validate(user) for user in db_users]

    @staticmethod
    def update_test_user_specification(user_id: int, specification: TestUserSpecification) -> None:
        with db.create_session() as session:
            test_user = session.get(db.TestUser, user_id)
            if test_user is None:
                raise TestUserDoesNotExistError(f"Test user with id {user_id} does not exist!")
            if not test_user.allow_update:
                raise TestUserUpdatingNotAllowedError(f"Test user updating with id {user_id} not allowed!")
            session.execute(
                sa.update(db.TestUser)
                .where(db.TestUser.id == test_user.id)
                .values(specification=specification, changed_at=get_current_time())
            )

    @staticmethod
    def delete_test_user(user_id: int) -> None:
        with db.create_session() as session:
            user = session.get(db.TestUser, user_id)
            if user is None:
                raise TestUserDoesNotExistError(f"Test user with id {user_id} does not exist!")
            session.delete(user)



import logging
from http import HTTPStatus

import fastapi

from overhave.api.deps import get_feature_tag_storage
from overhave.storage import IFeatureTagStorage, TagModel

logger = logging.getLogger(__name__)


def tags_item_handler(
    value: str,
    feature_tag_storage: IFeatureTagStorage = fastapi.Depends(get_feature_tag_storage),
) -> TagModel:
    logger.info("Getting %s with value='%s...'", TagModel.__name__, value)
    tag_model = feature_tag_storage.get_tag_by_value(value)
    if tag_model is None:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"Tag with value='{value}' does not exist"
        )
    return tag_model


def tags_list_handler(
    value: str,
    feature_tag_storage: IFeatureTagStorage = fastapi.Depends(get_feature_tag_storage),
) -> list[TagModel]:
    logger.info("Getting %s list like value='%s...'", TagModel.__name__, value)
    return feature_tag_storage.get_tags_like_value(value)



import logging

from pydantic import SecretStr

from overhave import db
from overhave.entities.auth_managers.secret_mixin import AdminSecretMixin
from overhave.storage import SystemUserModel

logger = logging.getLogger(__name__)


class DefaultAdminAuthorizationManager(AdminSecretMixin):
    """Class for user auth_managers.

    Manager authorize users with registered credentials.
    """

    def authorize_user(self, username: str, password: SecretStr) -> SystemUserModel | None:
        with db.create_session() as session:
            user = self._system_user_storage.get_user_by_credits(session=session, login=username, password=password)
            if user is None:
                return None
            return SystemUserModel.model_validate(user)
